import os
import csv
import json
from typing import List, Tuple, Dict

from openai import AsyncOpenAI
from openai.types.chat import ChatCompletionMessageParam
from air import login

auth = login(
    account=str(os.getenv("ACCOUNT")),
    api_key=str(os.getenv("API_KEY")),
    oauth_server=os.getenv("OAUTH_SERVER", ""),
)
base_url = os.getenv("AIREFINERY_ADDRESS", "")


async def simple_agent(query: str) -> str | None:
    """Create a simple agent using OpenAI chat completion."""
    global auth
    prompt = f"""Your task is to generate some synthetic data so that it will be useful to answer the user question. Do not mention this is synthetic data in your answer.\n\n{query}"""
    client = AsyncOpenAI(**auth.openai(base_url=base_url))
    response = await client.chat.completions.create(
        messages=[{"role": "user", "content": prompt}],
        model="meta-llama/Llama-3.1-70B-Instruct",
    )
    return response.choices[0].message.content


def create_evaluation_messages(
    query: str, response: str, criteria: List[Dict]
) -> List[ChatCompletionMessageParam]:
    """Create the evaluation message using provided evaluation criteria."""
    evaluation_template = {
        "evaluations": {
            criterion["name"]: {"score": 0, "explanation": ""} for criterion in criteria
        }
    }

    prompt = f"""
    Evaluate the following system output for the given query.

    Query: {query}

    System Output: {response}

    # Evaluation Criteria:
    # {json.dumps(criteria, indent=2)}

    Provide your evaluation in a JSON format with the following structure:

    {json.dumps(evaluation_template, indent=2)}

    Replace 'score' with a numeric rating and 'explanation' with justification for that score.
    Only return the JSON object as the output.
    """
    return [{"role": "user", "content": prompt}]


def save_to_csv(file_path: str, headers: List[str], rows: List[List]) -> None:
    """Saves data to a CSV file."""
    with open(file_path, "w", newline="") as csvfile:
        csv_writer = csv.writer(csvfile)
        csv_writer.writerow(headers)
        csv_writer.writerows(rows)


def flatten_evaluations(
    evaluations: Dict, criteria: List[Dict]
) -> Tuple[List[str], List]:
    """Flattens evaluation results into a list of scores and explanations."""
    criteria_names = [criterion["name"] for criterion in criteria]
    evaluation_header = [
        header
        for criterion in criteria_names
        for header in (f"{criterion}_score", f"{criterion}_explanation")
    ]
    flatten_values = []
    if evaluations and isinstance(evaluations, dict):
        for criterion in criteria_names:
            criterion_data = evaluations.get(criterion, {})
            flatten_values.extend(
                [
                    criterion_data.get("score", 0),
                    criterion_data.get("explanation", "No explanation provided"),
                ]
            )
    else:
        flatten_values = [
            0 if "_score" in header else "No explanation provided"
            for header in evaluation_header
        ]
    return evaluation_header, flatten_values


async def evaluate_response(
    query: str, result: str, evaluator_model: str, criteria: List[Dict]
) -> Tuple[List[str], List]:
    """Evaluate response using the evaluator model."""
    eval_messages = create_evaluation_messages(query, result, criteria)
    client = AsyncOpenAI(**auth.openai(base_url=base_url))
    eval_response = await client.chat.completions.create(
        messages=eval_messages, model=evaluator_model
    )

    # Extract and process evaluations
    eval_response = eval_response.choices[0].message.content
    if eval_response and isinstance(eval_response, str):
        eval_json = eval_response[
            eval_response.find("{") : eval_response.rfind("}") + 1
        ]
    else:
        eval_json = "{}"
    evaluations = json.loads(eval_json).get("evaluations", {})
    return flatten_evaluations(evaluations, criteria)
