import pytest
import pandas as pd
from unittest.mock import patch

from air.api import PandasAPI


def mock_read(file_type, data):
    """Helper function to mock pandas read methods for different file types."""
    if file_type == "csv":
        return pd.DataFrame(data)
    elif file_type == "parquet":
        return pd.DataFrame(data)


@pytest.mark.parametrize(
    "file_type, data, expected_keys",
    [
        ("csv", {"A": [1, 2], "B": [3, 4]}, ["default_test"]),
        ("parquet", {"C": [5, 6], "D": [7, 8]}, ["default_test"]),
    ],
)
def test_pandas_api_init(file_type, data, expected_keys):
    """Test the initialization of PandasAPI with CSV and Parquet files."""
    with patch(f"pandas.read_{file_type}", return_value=mock_read(file_type, data)):
        df_metadata_list = [{"name": "test", "file_path": f"test.{file_type}"}]
        client = PandasAPI(df_metadata_list)
        assert list(client.dataframes.keys()) == expected_keys
        assert client.dataframes["default_test"].equals(pd.DataFrame(data))


def test_pandas_api_unsupported_file_type(capfd):
    """Test handling of unsupported file types during PandasAPI initialization."""
    df_metadata_list = [{"name": "test", "file_path": "test.txt"}]
    client = PandasAPI(df_metadata_list)
    captured = capfd.readouterr()
    assert "Unsupported file type: .txt" in captured.out


@pytest.mark.asyncio
async def test_execute_query():
    """Test execute_query method with a valid query that extracts a DataFrame column."""
    data = {"A": [1, 2], "B": [3, 4]}
    df_metadata_list = [{"name": "test", "file_path": "test.csv"}]
    with patch("pandas.read_csv", return_value=pd.DataFrame(data)):
        client = PandasAPI(df_metadata_list)
        query = "default_test['A']"
        result, success = await client.execute_query(query)
        assert success is True
        assert result == [(1,), (2,)]


@pytest.mark.asyncio
async def test_execute_query_invalid():
    """Test execute_query method with an invalid query, accessing a non-existent column."""
    data = {"A": [1, 2], "B": [3, 4]}
    df_metadata_list = [{"name": "test", "file_path": "test.csv"}]
    with patch("pandas.read_csv", return_value=pd.DataFrame(data)):
        client = PandasAPI(df_metadata_list)
        query = "default_test['C']"  # Invalid column
        result, success = await client.execute_query(query)
        assert success is False
        assert isinstance(result, str)


@pytest.mark.asyncio
async def test_execute_query_scalar():
    """Test execute_query method with a query returning a scalar value."""
    data = {"A": [1, 2], "B": [3, 4]}
    df_metadata_list = [{"name": "test", "file_path": "test.csv"}]
    with patch("pandas.read_csv", return_value=pd.DataFrame(data)):
        client = PandasAPI(df_metadata_list)
        query = "default_test['A'].sum()"  # Scalar value query
        result, success = await client.execute_query(query)
        assert success is True
        assert result == [(3,)]
