"""
Test suite for the PostgresAPI class, ensuring successful and failed query executions.
"""

import logging
from typing import Any, Dict, Optional
from unittest.mock import patch

import pytest

from air.api.postgres import PostgresAPI


# Configure logging for testing
logging.basicConfig(level=logging.ERROR)


class Record:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __getitem__(self, key):
        return getattr(self, key)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()


class MockAsyncpgConnection:
    """Simulates an asyncpg connection for testing."""

    def __init__(self, expected_results=None):
        self.expected_results = expected_results or []
        self.execute_called_with = None

    async def fetch(self, query, *params):
        """Simulate executing a query and returning results."""
        self.execute_called_with = (query, params)
        return self.expected_results

    async def close(self):
        """Simulate closing the connection."""
        pass


# Mock connect function
async def mock_connect(*_args, **_kwargs):
    """Simulate connecting to the database and return a mock connection."""
    # Provide expected results for the test
    # expected_results = [{"row1": ("row1_col1", "row1_col2")}, {"row2": ("row2_col1", "row2_col2")}]
    # expected_results = [MockRecord(**dict(zip(["column1", "column2"], row))) for row in expected_results]
    expected_results = [("row1_col1", "row1_col2"), ("row2_col1", "row2_col2")]
    # Record = namedtuple('Record', ["col1", "col2"])
    column_names = ["col1", "col2"]
    expected_results = [
        Record(**dict(zip(column_names, row))) for row in expected_results
    ]
    return MockAsyncpgConnection(expected_results=expected_results)


@pytest.fixture
def mock_db_config() -> Dict[str, Any]:
    """
    Fixture to mock database configuration.
    """
    return {
        "host": "localhost",
        "port": 5432,
        "database": "test_db",
        "user": "test_user",
        "password": "test_password",
    }


@pytest.fixture
def postgres_api(mock_db_config: Dict[str, Any]) -> PostgresAPI:
    """
    Fixture to create an instance of PostgresAPI with the mock configuration.
    """
    return PostgresAPI(mock_db_config)


@pytest.mark.asyncio
async def test_postgres_api_initialization(mock_db_config: Dict[str, Any]) -> None:
    """
    Test case for PostgresAPI initialization.
    """
    client = PostgresAPI(mock_db_config)
    assert client.config == mock_db_config


@pytest.mark.asyncio
async def test_execute_query_success(postgres_api: PostgresAPI) -> None:
    """Test executing a query successfully with expected results."""
    query = "SELECT * FROM test_table WHERE id = 1"
    params: Optional[Any] = ()
    expected_results = [("row1_col1", "row1_col2"), ("row2_col1", "row2_col2")]
    column_names = ["col1", "col2"]
    mock_connection = MockAsyncpgConnection(
        expected_results=[
            Record(**dict(zip(column_names, row))) for row in expected_results
        ]
    )

    with patch("asyncpg.connect", return_value=mock_connection):
        # Execute the function
        result, success = await postgres_api.execute_query(query, params)

        # Assertions
        assert success is True
        assert result == expected_results

        # Ensure execute was called with the correct arguments
        assert mock_connection.execute_called_with == (query, params)


@pytest.mark.asyncio
async def test_execute_query_no_results(postgres_api: PostgresAPI) -> None:
    """Test executing a query that returns no results."""
    query = "SELECT * FROM empty_table"
    params: Optional[Any] = ()
    expected_results = []
    mock_connection = MockAsyncpgConnection(expected_results=expected_results)

    with patch("asyncpg.connect", return_value=mock_connection):
        result, success = await postgres_api.execute_query(query, params)

        assert success is True
        assert result == []

        assert mock_connection.execute_called_with == (query, params)


@pytest.mark.asyncio
async def test_execute_query_non_select(postgres_api: PostgresAPI) -> None:
    """Test executing a non-SELECT query."""
    query = "INSERT INTO test_table (id, name) VALUES (1, 'test')"
    params: Optional[Any] = ()
    expected_results = []
    mock_connection = MockAsyncpgConnection(expected_results=expected_results)

    with patch("asyncpg.connect", return_value=mock_connection):
        result, success = await postgres_api.execute_query(query, params)

        assert success is True
        assert result == []

        assert mock_connection.execute_called_with == (query, params)


@pytest.mark.asyncio
async def test_execute_query_failure(postgres_api: PostgresAPI) -> None:
    """Test handling a failure in executing a query due to connection error."""
    query = "SELECT * FROM test_table WHERE id = 1"

    params: Optional[Any] = ()
    mock_connection = MockAsyncpgConnection(
        expected_results=ConnectionError("Connection error")
    )

    with patch("asyncpg.connect", return_value=mock_connection):

        result, success = await postgres_api.execute_query(query, params)

        assert success is False
        assert isinstance(result, str)


@pytest.mark.asyncio
async def test_execute_query_invalid_query(postgres_api: PostgresAPI) -> None:
    """Test handling an invalid query that raises an exception."""
    query = ""  # Invalid empty query
    params: Optional[Any] = ()

    class MockAsyncpgConnectionWithException:
        """Simulates an asyncpg connection for testing."""

        def __init__(self, expected_results=None):
            self.expected_results = expected_results or []
            self.execute_called_with = None

        async def fetch(self, query, *params):
            """Simulate executing a query and returning results."""
            self.execute_called_with = (query, params)
            raise ValueError("Invalid query")

        async def close(self):
            """Simulate closing the connection."""
            pass

    mock_connection = MockAsyncpgConnectionWithException()

    with patch("asyncpg.connect", return_value=mock_connection):
        result, success = await postgres_api.execute_query(query, params)

        assert success is False
        assert isinstance(result, str)

        # Ensure execute was called
        assert mock_connection.execute_called_with == (query, params)


@pytest.mark.asyncio
async def test_execute_query_with_params(postgres_api: PostgresAPI) -> None:
    """Test executing a query with parameters."""
    query = "SELECT * FROM test_table WHERE id = %s AND name = %s"
    params = (1, "test")
    expected_results = [("row1_col1", "row1_col2")]
    column_names = ["col1", "col2"]
    mock_connection = MockAsyncpgConnection(
        expected_results=[
            Record(**dict(zip(column_names, row))) for row in expected_results
        ]
    )

    with patch("asyncpg.connect", return_value=mock_connection):
        result, success = await postgres_api.execute_query(query, params)

        assert success is True
        assert result == expected_results

        assert mock_connection.execute_called_with == (query, params)


@pytest.mark.asyncio
async def test_execute_query_exception_handling(postgres_api: PostgresAPI) -> None:
    """Test handling an exception raised during query execution."""
    query = "SELECT * FROM test_table WHERE id = 1"
    params: Optional[Any] = ()

    # Mock cursor that raises an exception on execute
    class MockAsyncpgConnectionWithException:
        """Simulates an asyncpg connection for testing."""

        def __init__(self, expected_results=None):
            self.expected_results = expected_results or []
            self.execute_called_with = None

        async def fetch(self, query, *params):
            """Simulate executing a query and returning results."""
            self.execute_called_with = (query, params)
            raise RuntimeError("An error occured")

        async def close(self):
            """Simulate closing the connection."""
            pass

    mock_connection = MockAsyncpgConnectionWithException()

    with patch("asyncpg.connect", return_value=mock_connection):
        result, success = await postgres_api.execute_query(query, params)

        assert success is False
        assert isinstance(result, str)
        # Ensure execute was called
        assert mock_connection.execute_called_with == (query, params)


@pytest.mark.asyncio
async def test_execute_query_connection_error(postgres_api: PostgresAPI) -> None:
    """Test handling an exception raised during query execution."""
    query = "SELECT * FROM test_table WHERE id = 1"
    params: Optional[Any] = ()

    result, success = await postgres_api.execute_query(query, params)

    assert success is False
    assert isinstance(result, str)
