from unittest.mock import MagicMock, Mock, patch

import pytest
from requests.exceptions import HTTPError

from air.api.vector_db import AzureAISearch, BaseVectorDB, VectorDBConfig
from air.embeddings import EmbeddingsClient


@pytest.fixture
def mock_vectordb_config() -> VectorDBConfig:
    """
    Fixture to mock vector DB configuration
    """
    return VectorDBConfig(
        base_url="http://test-url.com",
        api_key="test-key",
        api_version="test-version",
        index="test-vectordb",
        embedding_column="text_vector",
        top_k=1,
        content_column=["text"],
    )


@pytest.fixture
def mock_azure_aisearch(request) -> AzureAISearch:
    """
    Fixture to mock AzureAISearch class
    """
    vectordb_config = request.getfixturevalue("mock_vectordb_config")
    return AzureAISearch(vectordb_config)


def test_init(request):
    """
    Function to test the initialization of AzureAISearch
    """
    vectordb_config = request.getfixturevalue("mock_vectordb_config")
    azure_aisearch = AzureAISearch(vectordb_config)
    assert isinstance(azure_aisearch, BaseVectorDB)
    assert (
        azure_aisearch.index_url
        == f"{vectordb_config.base_url}/indexes/{vectordb_config.index}/docs/index?api-version={vectordb_config.api_version}"
    )
    assert (
        azure_aisearch.search_url
        == f"{vectordb_config.base_url}/indexes/{vectordb_config.index}/docs/search?api-version={vectordb_config.api_version}"
    )
    assert azure_aisearch.fields == "text_vector"
    assert azure_aisearch.k == 1
    assert azure_aisearch.select == "text"
    assert azure_aisearch.timeout == 60
    assert azure_aisearch.headers == {
        "Content-Type": "application/json",
        "api-key": "test-key",
    }


@patch("requests.post")
def test_upload(mock_post, request):
    """
    Function to test the upload method of AzureAISearch
    """
    mock_response = Mock()
    mock_response.status_code = 200
    mock_response.json.return_value = {"status": "success"}
    mock_post.return_value = mock_response

    data = [
        {
            "id": i,
            "text": f"Hello number {i}",
            "text_vector": [i * 0.1, i * 0.2, i * 0.3],
        }
        for i in range(20)
    ]
    azure_aisearch = request.getfixturevalue("mock_azure_aisearch")
    result = azure_aisearch.upload(data)
    actual_data = [dict(row, **{"@search.action": "upload"}) for row in data]
    assert result is True
    mock_post.assert_called_once_with(
        azure_aisearch.index_url,
        headers=azure_aisearch.headers,
        json={"value": actual_data},
        timeout=60,
    )


@patch("air.api.vector_db.azure_aisearch.EmbeddingsClient.create")
@patch("air.api.vector_db.azure_aisearch.EmbeddingsClient")
@patch("air.api.vector_db.azure_aisearch.auth")
def test_get_query_embedding(mock_auth, mock_emb_client, mock_emb_create, request):
    """
    Function to test the response of get_query_embedding function from AzureAISearch
    """
    mock_auth.account = "test-account"
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_embedding = MagicMock()
    mock_embedding.embedding = [0.1, 0.2, 0.3]
    mock_response.data = [mock_embedding]
    mock_emb_create.return_value = mock_response
    azure_aisearch = request.getfixturevalue("mock_azure_aisearch")
    response = azure_aisearch.get_query_embedding(
        query="test-query",
        embedding_client=mock_emb_client,
        embedding_model="test-embedding-model",
    )
    assert response == [0.1, 0.2, 0.3]


@patch("air.api.vector_db.azure_aisearch.EmbeddingsClient.create")
@patch("air.api.vector_db.azure_aisearch.EmbeddingsClient")
@patch("air.api.vector_db.azure_aisearch.auth")
def test_get_query_embedding_failure(
    mock_auth, mock_emb_client, mock_emb_create, request
):
    """
    Function to test the response of get_query_embedding function from AzureAISearch
    when the client raises a HTTPError
    """
    mock_auth.account = "test-account"
    mock_emb_create.side_effect = HTTPError("Invalid input")
    azure_aisearch = request.getfixturevalue("mock_azure_aisearch")
    response = azure_aisearch.get_query_embedding(
        query="test-query",
        embedding_client=mock_emb_client,
        embedding_model="test-embedding-model",
    )
    assert response == []


@patch("air.api.vector_db.azure_aisearch.AzureAISearch.get_query_embedding")
@patch("requests.post")
def test_vector_search_invalid_json_recieved(mock_post, mock_embedding, request):
    """
    Function to test vector search when invalid response is returned from
    index search
    """

    # Mock the embedding
    mock_embedding.return_value = [0.1, 0.2, 0.3]

    # Mock the API response
    mock_post.return_value.text = (
        '{"value": [{"text": "text1", {"text": "text2"}]}'  # invalid json
    )
    azure_aisearch = request.getfixturevalue("mock_azure_aisearch")
    result = azure_aisearch.vector_search(
        "test query",
        EmbeddingsClient(base_url="test-base-url", api_key="test-key"),
        "test-embedding-model",
    )

    # Assertions. Should return an empty list
    assert result == []


@patch("air.api.vector_db.azure_aisearch.AzureAISearch.get_query_embedding")
@patch("requests.post")
def test_vector_search(mock_post, mock_embedding, request):
    """
    Function to test vector search
    """

    # Mock the embedding
    mock_embedding.return_value = [0.1, 0.2, 0.3]

    # Mock the API response
    mock_post.return_value.text = '{"value": [{"text": "text1"}, {"text": "text2"}]}'

    azure_aisearch = request.getfixturevalue("mock_azure_aisearch")
    result = azure_aisearch.vector_search(
        "test query",
        EmbeddingsClient(base_url="test-base-url", api_key="test-api-key"),
        "test-embedding-model",
    )

    # Assertions
    assert len(result) == 2
    assert result[0]["text"] == "text1"
    assert result[1]["text"] == "text2"
