# test_analytics_executor.py
# pylint: disable=unused-import, too-many-positional-arguments, unused-argument, duplicate-code

import asyncio
from unittest.mock import patch

import pytest

from air.distiller.executor.analytics_executor import AnalyticsExecutor


# Mock implementations of PostgresAPI and PandasAPI
class MockPostgresAPI:
    """Mock implementation of PostgresAPI for testing purposes."""

    def __init__(self, db_config):
        self.db_config = db_config

    async def execute_query(self, query, **kwargs):
        return f"Executed Postgres query: {query} with config {self.db_config}"


class MockPandasAPI:
    """Mock implementation of PandasAPI for testing purposes."""

    def __init__(self, tables):
        self.tables = tables

    async def execute_query(self, query, **kwargs):
        return f"Executed Pandas query: {query} on tables {self.tables}"


# Mock the import of PostgresAPI and PandasAPI in the module where AnalyticsExecutor is defined
# We'll use patch to replace these with our mocks during the tests
@pytest.fixture
def mock_postgres_api():
    """Fixture to mock the PostgresAPI class with MockPostgresAPI."""
    with patch(
        "air.distiller.executor.analytics_executor.PostgresAPI", new=MockPostgresAPI
    ):
        yield


@pytest.fixture
def mock_pandas_api():
    """Fixture to mock the PandasAPI class with MockPandasAPI."""
    with patch(
        "air.distiller.executor.analytics_executor.PandasAPI", new=MockPandasAPI
    ):
        yield


# Now we can write tests


@pytest.mark.asyncio
async def test_analytics_executor_init_postgres_valid(mock_postgres_api):
    """
    Test initializing AnalyticsExecutor with valid Postgres executor_config.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "executor_config": {
            "type": "PostgresExecutor",
            "db_config": {"host": "localhost", "port": 5432, "dbname": "testdb"},
        }
    }

    executor = AnalyticsExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    # Check that "PostgresExecutor" has been added to executor.func
    assert "PostgresExecutor" in executor.func
    # Test that it's callable
    result = await executor.func["PostgresExecutor"]("SELECT * FROM table;")
    expected_result = (
        "Executed Postgres query: SELECT * FROM table; with config "
        "{'host': 'localhost', 'port': 5432, 'dbname': 'testdb'}"
    )
    assert result == expected_result


@pytest.mark.asyncio
async def test_analytics_executor_init_pandas_valid(mock_pandas_api):
    """
    Test initializing AnalyticsExecutor with valid Pandas executor_config.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "executor_config": {"type": "PandasExecutor", "tables": ["table1", "table2"]}
    }

    executor = AnalyticsExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    # Check that "PandasExecutor" has been added to executor.func
    assert "PandasExecutor" in executor.func

    # Test that it's callable
    result = await executor.func["PandasExecutor"]("SELECT * FROM table;")
    expected_result = (
        "Executed Pandas query: SELECT * FROM table; on tables ['table1', 'table2']"
    )
    assert result == expected_result


@pytest.mark.asyncio
async def test_analytics_executor_call_with_valid_executor(mock_postgres_api):
    """
    Test that AnalyticsExecutor can execute the correct function based on __executor__.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "executor_config": {
            "type": "PostgresExecutor",
            "db_config": {"host": "localhost", "port": 5432, "dbname": "testdb"},
        }
    }

    executor = AnalyticsExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    # Now call the executor with query as a positional argument
    result = await executor(
        "test_id",
        query="SELECT * FROM users;",
        __executor__="PostgresExecutor",
    )

    expected_result = (
        "Executed Postgres query: SELECT * FROM users; with config "
        "{'host': 'localhost', 'port': 5432, 'dbname': 'testdb'}"
    )
    assert result == expected_result

    # Check that send_queue has received the message
    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == expected_result

    # Now call the executor
    result = await executor(
        request_id="test_id",
        query="SELECT * FROM users;",
        __executor__="PostgresExecutor",
    )


@pytest.mark.asyncio
async def test_analytics_executor_call_with_invalid_executor(mock_postgres_api):
    """
    Test that AnalyticsExecutor raises error when using invalid __executor__.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "executor_config": {
            "type": "PostgresExecutor",
            "db_config": {"host": "localhost", "port": 5432, "dbname": "testdb"},
        }
    }

    executor = AnalyticsExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    # Try calling with invalid __executor__
    with pytest.raises(ValueError) as excinfo:
        await executor(
            request_id="test_id",
            query="SELECT * FROM users;",
            __executor__="InvalidExecutor",
        )
    assert "Retriever type 'InvalidExecutor' is not available in func." in str(
        excinfo.value
    )


@pytest.mark.asyncio
async def test_analytics_executor_init_missing_executor_config():
    """
    Test that AnalyticsExecutor raises error when executor_config is missing.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        # No executor_config
    }

    with pytest.raises(ValueError) as excinfo:
        AnalyticsExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert "executor_config is missing in utility_config." in str(excinfo.value)


@pytest.mark.asyncio
async def test_analytics_executor_init_missing_type_in_executor_config():
    """
    Test that AnalyticsExecutor raises error when executor_config.type is missing.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "executor_config": {
            # Missing "type"
            "db_config": {"host": "localhost", "port": 5432, "dbname": "testdb"}
        }
    }

    with pytest.raises(ValueError) as excinfo:
        AnalyticsExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert "executor_config.type is missing in utility_config." in str(excinfo.value)


@pytest.mark.asyncio
async def test_analytics_executor_init_unsupported_executor_type():
    """
    Test that AnalyticsExecutor raises error when retriever type is unsupported.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "executor_config": {
            "type": "UnsupportedType",
            "db_config": {"host": "localhost", "port": 5432, "dbname": "testdb"},
        }
    }

    with pytest.raises(ValueError) as excinfo:
        AnalyticsExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert "Unsupported executor type: UnsupportedType." in str(excinfo.value)


@pytest.mark.asyncio
async def test_analytics_executor_postgres_missing_db_config():
    """
    Test that AnalyticsExecutor raises error when db_config is missing for Postgres.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "executor_config": {
            "type": "PostgresExecutor",
            # Missing "db_config"
        }
    }

    with pytest.raises(ValueError) as excinfo:
        AnalyticsExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert "db_config is missing in executor_config." in str(excinfo.value)


@pytest.mark.asyncio
async def test_analytics_executor_pandas_missing_tables():
    """
    Test that AnalyticsExecutor raises error when tables are missing for PandasExecutor.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "executor_config": {
            "type": "PandasExecutor",
            # Missing "tables"
        }
    }

    with pytest.raises(ValueError) as excinfo:
        AnalyticsExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert "tables is missing in executor_config." in str(excinfo.value)


@pytest.mark.asyncio
async def test_analytics_executor_uses_user_defined_postgres_executor(
    mock_postgres_api,
):
    """
    Test that AnalyticsExecutor uses user-defined PostgresExecutor if provided.
    """
    send_queue = asyncio.Queue()

    # User-defined PostgresExecutor function
    def user_defined_postgres_executor(query):
        """User-defined executor function."""
        return f"User-defined executor executed query: {query}"

    func_dict = {"PostgresExecutor": user_defined_postgres_executor}
    utility_config = {
        "executor_config": {
            "type": "PostgresExecutor",
            "db_config": {"host": "localhost", "port": 5432, "dbname": "testdb"},
        }
    }

    executor = AnalyticsExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    # Ensure that the user-defined function is used
    result = await executor(
        request_id="test_id",
        query="SELECT * FROM users;",
        __executor__="PostgresExecutor",
    )

    expected_result = "User-defined executor executed query: SELECT * FROM users;"
    assert result == expected_result

    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == expected_result
