import asyncio
import sys
import types
from unittest.mock import MagicMock, patch

import pytest

# ────── Stub Azure packages so azure_executor can import cleanly ──────  # pragma: no cover
if "azure" not in sys.modules:
    sys.modules["azure"] = types.ModuleType("azure")
if "azure.ai" not in sys.modules:
    sys.modules["azure.ai"] = types.ModuleType("azure.ai")
proj_mod = types.ModuleType("azure.ai.projects")
setattr(proj_mod, "AIProjectClient", MagicMock())
sys.modules["azure.ai.projects"] = proj_mod
iden_mod = types.ModuleType("azure.identity")
setattr(iden_mod, "DefaultAzureCredential", MagicMock())
sys.modules["azure.identity"] = iden_mod

from air.distiller.executor.azure_executor import AzureExecutor


@pytest.fixture
def send_queue():
    """Fixture providing an asyncio queue for tests."""
    return asyncio.Queue()


@pytest.fixture
def mock_project_client():
    """Fixture that returns a mocked Azure AI project client."""
    mock_pc = MagicMock()
    # Create a fake agent with an id attribute.
    fake_agent = MagicMock()
    fake_agent.id = "test_agent_id"
    mock_pc.agents.get_agent.return_value = fake_agent
    return mock_pc


def test_init_success(send_queue, mock_project_client):
    """Test successful initialization of AzureExecutor."""
    utility_config = {"agent_id": "test_agent", "connection_string": "fake_connection"}
    with patch(
        "azure.ai.projects.AIProjectClient.from_connection_string",
        return_value=mock_project_client,
    ) as mock_from_conn:
        executor = AzureExecutor(
            func={},  # Unused in this test.
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )

        # Verify that the project client was set correctly and the agent was fetched.
        assert executor.project_client is mock_project_client
        mock_project_client.agents.get_agent.assert_called_once_with("test_agent")
        mock_from_conn.assert_called_once()


def test_init_missing_agent_id(send_queue):
    """Test that initialization fails if 'agent_id' is missing."""
    utility_config = {"connection_string": "fake_connection"}
    with pytest.raises(ValueError) as exc_info:
        AzureExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert "Missing 'agent_id'" in str(exc_info.value)


def test_init_missing_connection_string(send_queue):
    """Test that initialization fails if 'connection_string' is missing."""
    utility_config = {"agent_id": "test_agent"}
    with pytest.raises(ValueError) as exc_info:
        AzureExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert "Missing 'connection_string'" in str(exc_info.value)


def test_execute_agent_no_prompt(send_queue, mock_project_client):
    """Test that _execute_agent raises ValueError when no prompt is provided."""
    utility_config = {"agent_id": "test_agent", "connection_string": "fake_connection"}
    with patch(
        "azure.ai.projects.AIProjectClient.from_connection_string",
        return_value=mock_project_client,
    ):
        executor = AzureExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )

        with pytest.raises(ValueError) as exc_info:
            executor._execute_agent()
        assert "Missing 'prompt' parameter in AzureExecutor._execute_agent." in str(
            exc_info.value
        )


def test_execute_agent_success(send_queue, mock_project_client):
    """Test that _execute_agent successfully returns the agent's string response."""
    utility_config = {"agent_id": "test_agent", "connection_string": "fake_connection"}
    with patch(
        "azure.ai.projects.AIProjectClient.from_connection_string",
        return_value=mock_project_client,
    ):
        executor = AzureExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )

        # Simulate asynchronous _run_agent behavior.
        fake_thread = MagicMock()
        fake_thread.id = "test_thread"
        mock_project_client.agents.create_thread.return_value = fake_thread

        # Simulate posting a message and processing a run (no return value required).
        # Simulate list_messages returning a single assistant message.
        fake_message = {
            "role": "assistant",
            "content": [{"type": "text", "text": {"value": "Hello from Azure!"}}],
        }
        mock_project_client.agents.list_messages.return_value = {"data": [fake_message]}

        # Execute _execute_agent with a valid prompt.
        result = executor._execute_agent(prompt="Hello Agent")
        assert result == "Hello from Azure!"

        # Verify that the thread creation, message posting, run processing,
        # message listing, and thread deletion were called.
        mock_project_client.agents.create_thread.assert_called_once()
        mock_project_client.agents.create_message.assert_called_once()
        mock_project_client.agents.create_and_process_run.assert_called_once()
        mock_project_client.agents.list_messages.assert_called_once()
        mock_project_client.agents.delete_thread.assert_called_once_with("test_thread")


@patch("azure.ai.projects.AIProjectClient.from_connection_string")
def test_init_client_failure(mock_from_conn, send_queue):
    """Test handling of exceptions during client initialization."""
    mock_from_conn.side_effect = Exception("Connection failed")
    utility_config = {"agent_id": "test_agent", "connection_string": "fake_connection"}

    with pytest.raises(Exception) as exc_info:
        AzureExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
        )

    assert "Connection failed" in str(exc_info.value)


def test_run_agent_thread_creation_failure(send_queue, mock_project_client):
    """Test exception handling when thread creation fails."""
    utility_config = {"agent_id": "test_agent", "connection_string": "fake_connection"}

    with patch(
        "azure.ai.projects.AIProjectClient.from_connection_string",
        return_value=mock_project_client,
    ):
        # Setup mock to raise exception when creating thread
        mock_project_client.agents.create_thread.side_effect = Exception(
            "Thread creation failed"
        )

        executor = AzureExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
        )

        with pytest.raises(Exception) as exc_info:
            executor._execute_agent(prompt="Test prompt")

        assert "Thread creation failed" in str(exc_info.value)


def test_run_agent_create_message_failure(send_queue, mock_project_client):
    """Test exception handling when message creation fails."""
    utility_config = {"agent_id": "test_agent", "connection_string": "fake_connection"}

    with patch(
        "azure.ai.projects.AIProjectClient.from_connection_string",
        return_value=mock_project_client,
    ):
        # Setup mock for thread creation success but message post failure
        fake_thread = MagicMock()
        fake_thread.id = "test_thread"
        mock_project_client.agents.create_thread.return_value = fake_thread
        mock_project_client.agents.create_message.side_effect = Exception(
            "Message creation failed"
        )

        executor = AzureExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
        )

        with pytest.raises(Exception) as exc_info:
            executor._execute_agent(prompt="Test prompt")

        assert "Message creation failed" in str(exc_info.value)


def test_run_agent_process_run_failure(send_queue, mock_project_client):
    """Test exception handling when processing a run fails."""
    utility_config = {"agent_id": "test_agent", "connection_string": "fake_connection"}

    with patch(
        "azure.ai.projects.AIProjectClient.from_connection_string",
        return_value=mock_project_client,
    ):
        # Setup mocks for successful thread and message creation but run processing failure
        fake_thread = MagicMock()
        fake_thread.id = "test_thread"
        mock_project_client.agents.create_thread.return_value = fake_thread
        mock_project_client.agents.create_and_process_run.side_effect = Exception(
            "Run processing failed"
        )

        executor = AzureExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
        )

        with pytest.raises(Exception) as exc_info:
            executor._execute_agent(prompt="Test prompt")

        assert "Run processing failed" in str(exc_info.value)


def test_run_agent_list_messages_failure(send_queue, mock_project_client):
    """Test exception handling when retrieving messages fails."""
    utility_config = {"agent_id": "test_agent", "connection_string": "fake_connection"}

    with patch(
        "azure.ai.projects.AIProjectClient.from_connection_string",
        return_value=mock_project_client,
    ):
        # Setup mocks for successful operations but message retrieval failure
        fake_thread = MagicMock()
        fake_thread.id = "test_thread"
        mock_project_client.agents.create_thread.return_value = fake_thread
        mock_project_client.agents.list_messages.side_effect = Exception(
            "Message retrieval failed"
        )

        executor = AzureExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
        )

        with pytest.raises(Exception) as exc_info:
            executor._execute_agent(prompt="Test prompt")

        assert "Message retrieval failed" in str(exc_info.value)


def test_run_agent_thread_deletion_failure(send_queue, mock_project_client):
    """Test that execution completes successfully even if thread deletion fails."""
    utility_config = {"agent_id": "test_agent", "connection_string": "fake_connection"}

    with patch(
        "azure.ai.projects.AIProjectClient.from_connection_string",
        return_value=mock_project_client,
    ):
        # Setup thread creation, successful message processing, but thread deletion failure
        fake_thread = MagicMock()
        fake_thread.id = "test_thread"
        mock_project_client.agents.create_thread.return_value = fake_thread

        # Successful message list return
        fake_message = {
            "role": "assistant",
            "content": [{"type": "text", "text": {"value": "Test response"}}],
        }
        mock_project_client.agents.list_messages.return_value = {"data": [fake_message]}

        # Thread deletion raises exception
        mock_project_client.agents.delete_thread.side_effect = Exception(
            "Thread deletion failed"
        )

        executor = AzureExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
        )

        # Even with thread deletion failure, the function should complete successfully
        result = executor._execute_agent(prompt="Test prompt")
        assert result == "Test response"

        # Verify delete_thread was called but the exception was handled
        mock_project_client.agents.delete_thread.assert_called_once_with("test_thread")


def test_run_agent_multiple_message_parts(send_queue, mock_project_client):
    """Test handling of multiple message parts from the assistant."""
    utility_config = {"agent_id": "test_agent", "connection_string": "fake_connection"}

    with patch(
        "azure.ai.projects.AIProjectClient.from_connection_string",
        return_value=mock_project_client,
    ):
        # Setup thread creation and return multiple message parts
        fake_thread = MagicMock()
        fake_thread.id = "test_thread"
        mock_project_client.agents.create_thread.return_value = fake_thread

        # Message with multiple content parts and different types
        fake_messages = {
            "data": [
                {
                    "role": "user",
                    "content": [{"type": "text", "text": {"value": "User question"}}],
                },
                {
                    "role": "assistant",
                    "content": [
                        {"type": "text", "text": {"value": "Part 1"}},
                        {"type": "text", "text": {"value": "Part 2"}},
                        {"type": "non_text", "data": {"value": "Should be ignored"}},
                        {"type": "text", "text": {}},  # Empty value
                    ],
                },
                {
                    "role": "assistant",
                    "content": [{"type": "text", "text": {"value": "Part 3"}}],
                },
            ]
        }
        mock_project_client.agents.list_messages.return_value = fake_messages

        executor = AzureExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
        )

        result = executor._execute_agent(prompt="Test prompt")
        # The result should concatenate all valid text parts
        assert result == "Part 1\nPart 2\nPart 3"
