# test_executor.py
# pylint: disable=unused-import, too-many-positional-arguments, unused-argument, duplicate-code
"""
Test module for Executor functionality.
Contains tests for async and blocking custom agents, cancellation, exceptions,
and handling of additional or missing arguments.
"""

import asyncio
import time
from typing import Callable

import pytest

# Import the Executor from your executor module
from air.distiller.executor.executor import Executor


def mock_custom_agent(query: str) -> str:
    """
    Mock function simulating a basic custom agent.
    """
    _ = query
    return "processed_query"


async def mock_async_custom_agent(query: str) -> str:
    """
    Mock function simulating an asynchronous custom agent.
    """
    _ = query
    await asyncio.sleep(0.1)
    return "async_processed_query"


def mock_slow_agent(query: str) -> str:
    """
    Mock function simulating a slow-responding agent.
    """
    _ = query
    time.sleep(1)  # Sleep to simulate a slow agent
    return "delayed_result"


def mock_exception_raising_agent(query: str) -> str:
    """
    Agent function that directly raises an exception.
    """
    _ = query
    raise ValueError("Simulated agent failure")


@pytest.mark.asyncio
async def test_executor_with_async_custom_agent():
    """
    Test that an async custom agent is handled correctly by the Executor.
    """
    send_queue = asyncio.Queue()
    query = "test_query"

    executor = Executor(
        func=mock_async_custom_agent,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        return_string=False,
    )

    # Execute the async function through the executor
    result = await executor(
        "test_id",  # request_id
        query,  # Pass the query as a positional argument
    )

    # Verify the result
    assert result == "async_processed_query"

    # Verify that the result was put into the send_queue
    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == "async_processed_query"


@pytest.mark.asyncio
async def test_executor_with_blocking_custom_agent():
    """
    Test that a blocking (synchronous) custom agent is handled correctly by the Executor.
    """
    send_queue = asyncio.Queue()
    query = "test_query"

    executor = Executor(
        func=mock_custom_agent,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        return_string=False,
    )

    # Execute the sync function through the executor
    result = await executor(
        "test_id",  # request_id
        query,  # Pass the query as a positional argument
    )

    # Verify the result
    assert result == "processed_query"

    # Verify that the result was put into the send_queue
    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == "processed_query"


@pytest.mark.asyncio
async def test_executor_worker_exception():
    """
    Test handling of an exception raised directly by the worker function.
    """
    send_queue = asyncio.Queue()
    query = "test_query"

    executor = Executor(
        func=mock_exception_raising_agent,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        return_string=False,
    )

    with pytest.raises(RuntimeError) as excinfo:
        await executor(
            "test_id",
            query,  # Pass in the query as a positional argument
        )

    # Verify the raised exception contains the agent's error message
    assert "Simulated agent failure" in str(excinfo.value)
    # Also verify we see the "Exception in function execution" prefix
    assert "Exception in function execution:" in str(excinfo.value)


@pytest.mark.asyncio
async def test_executor_cancelled():
    """
    Test handling of cancelled tasks.
    The Executor should handle cancellation gracefully.
    """
    send_queue = asyncio.Queue()
    query = "test_query"

    executor = Executor(
        func=mock_slow_agent,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        return_string=False,
    )

    async def run_with_cancel():
        """Helper function to run and then cancel the executor task."""
        task = asyncio.create_task(
            executor(
                "test_id",
                query,
            )
        )
        await asyncio.sleep(0.5)
        task.cancel()
        with pytest.raises(asyncio.CancelledError):
            await task

    await run_with_cancel()

    # Note: Since ThreadPoolExecutor threads cannot be easily cancelled,
    # the thread may continue running in the background.
    # We can optionally check if the send_queue remains empty,
    # but this is not guaranteed due to the nature of threads.
    # For this test, we can ensure our code handles the cancellation without errors.


@pytest.mark.asyncio
async def test_executor_with_additional_arguments():
    """
    Test that the Executor ignores extra arguments not accepted by the function.
    """
    send_queue = asyncio.Queue()

    def sample_func(a, b=2):
        return a + b

    executor = Executor(
        func=sample_func,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        return_string=True,
    )

    result = await executor(
        "test_id",
        1,
        2,
        3,
        4,
        5,  # Extra positional arguments (ignored)
        c=3,
        d=4,
        e=5,  # Extra keyword arguments (ignored)
    )

    # Verify the result
    assert result == 3  # 1 + 2 (int result)
    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == "3"  # String in send_queue


@pytest.mark.asyncio
async def test_executor_with_missing_required_argument():
    """
    Test that the Executor raises an exception if required positional arguments are missing.
    """
    send_queue = asyncio.Queue()

    def sample_func(a, b):
        return a + b

    executor = Executor(
        func=sample_func,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        return_string=True,
    )

    with pytest.raises(RuntimeError) as excinfo:
        await executor(
            "test_id",
            1,  # Missing 'b'
        )
    exc = excinfo.value
    # Check the original exception type
    assert isinstance(exc.__cause__, TypeError)
    # Optionally, check the message
    assert "missing 1 required positional argument" in str(exc)


# New tests to verify functions with *args and **kwargs


@pytest.mark.asyncio
async def test_executor_with_varargs_function():
    """
    Test that the Executor passes all positional arguments to functions accepting *args.
    """
    send_queue = asyncio.Queue()

    def sample_func(a, b=2, *args):
        return a + b + sum(args)

    executor = Executor(
        func=sample_func,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        return_string=True,
    )

    result = await executor(
        "test_id",
        1,  # a = 1
        2,  # b = 2
        3,
        4,
        5,  # *args = (3, 4, 5)
    )

    # Verify the result
    assert result == 1 + 2 + 3 + 4 + 5  # 15
    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == "15"  # String in send_queue


@pytest.mark.asyncio
async def test_executor_with_varkwargs_function():
    """
    Test that the Executor passes all keyword arguments to functions accepting **kwargs.
    """
    send_queue = asyncio.Queue()

    def sample_func(a, b=2, **kwargs):
        return a + b + sum(kwargs.values())

    executor = Executor(
        func=sample_func,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        return_string=True,
    )

    result = await executor(
        "test_id",
        1,  # a = 1
        b=2,  # b = 2
        c=3,
        d=4,
        e=5,  # **kwargs = {'c': 3, 'd': 4, 'e': 5}
    )

    # Verify the result
    assert result == 1 + 2 + 3 + 4 + 5  # 15
    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == "15"  # String in send_queue


@pytest.mark.asyncio
async def test_executor_with_varargs_and_varkwargs_function():
    """
    Test that the Executor passes all positional and keyword arguments to functions accepting *args and **kwargs.
    """
    send_queue = asyncio.Queue()

    def sample_func(a, b=2, *args, **kwargs):
        return a + b + sum(args) + sum(kwargs.values())

    executor = Executor(
        func=sample_func,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        return_string=True,
    )

    result = await executor(
        "test_id",
        1,  # a = 1
        2,  # b = 2
        3,
        4,  # *args = (3, 4)
        c=5,
        d=6,  # **kwargs = {'c': 5, 'd': 6}
    )

    # Verify the result
    assert result == 1 + 2 + 3 + 4 + 5 + 6  # 21
    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == "21"  # String in send_queue


@pytest.mark.asyncio
async def test_executor_with_unexpected_keyword_arguments():
    """
    Test that the Executor ignores unexpected keyword arguments when the function does not accept **kwargs.
    """
    send_queue = asyncio.Queue()

    def sample_func(a, b=2):
        return a + b

    executor = Executor(
        func=sample_func,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        return_string=True,
    )

    result = await executor(
        "test_id",
        1,
        b=2,
        c=3,  # Unexpected keyword argument (should be ignored)
    )

    # Verify the result
    assert result == 3  # 1 + 2
    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == "3"  # String in send_queue


@pytest.mark.asyncio
async def test_executor_with_unexpected_keyword_arguments_error():
    """
    Test that the Executor raises an exception when passing unexpected keyword arguments
    to a function that does not accept **kwargs, if argument filtering fails.
    """
    send_queue = asyncio.Queue()

    def sample_func(a, b=2):
        return a + b

    # Modify the executor to simulate a scenario where argument filtering fails.
    # For this test, we'll assume that the _filter_arguments method does not filter out
    # unexpected keyword arguments (e.g., accepts_varkwargs detection is incorrect).
    # Since our current implementation filters them out, we'll need to mock this behavior.

    # We'll create a subclass to simulate the failure.
    class FaultyExecutor(Executor):
        def _filter_arguments(self, func: Callable, args: tuple, kwargs: dict):
            # Intentionally pass all arguments without filtering to simulate the failure.
            return args, kwargs

    executor = FaultyExecutor(
        func=sample_func,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        return_string=True,
    )

    with pytest.raises(RuntimeError) as excinfo:
        await executor(
            "test_id",
            1,
            b=2,
            c=3,  # Unexpected keyword argument (should cause an error)
        )

    exc = excinfo.value
    # Check the original exception type
    assert isinstance(exc.__cause__, TypeError)
    # Optionally, check the message contains information about the unexpected keyword argument
    assert "got an unexpected keyword argument 'c'" in str(exc.__cause__)
