"""Unit tests for air.distiller.executor initialization code."""

import unittest
from unittest.mock import MagicMock, patch

# Import the functions we want to test:
from air.distiller.executor import (
    get_all_exeecutor_agents,
    get_executor,
    get_executor_class,
)


class ExecutorInitTest(unittest.TestCase):
    """Tests for the lazy-loading functions in air.distiller.executor."""

    def test_get_all_exeecutor_agents(self):
        """Tests that get_all_exeecutor_agents returns all known agent classes."""
        agents = get_all_exeecutor_agents()
        expected_agents = [
            "CustomAgent",
            "AnalyticsAgent",
            "ToolUseAgent",
            "GoogleAgent",
            "AzureAIAgent",
            "MCPClientAgent",
            "WriterAIAgent",
            "SalesforceAgent",
            "ResearchAgent",
        ]
        self.assertCountEqual(
            agents,
            expected_agents,
            "get_all_exeecutor_agents() did not return the expected agent list.",
        )

    @patch("importlib.import_module")
    def test_get_executor_class_valid(self, mock_import):
        """Tests that get_executor_class lazily imports and returns the correct class."""
        # Prepare a mock module/class.
        mock_module = MagicMock()
        mock_module.Executor = MagicMock()
        mock_import.return_value = mock_module

        # Call the function under test with a valid agent class.
        cls = get_executor_class("CustomAgent")

        # Ensure the correct module was imported and the class was retrieved.
        mock_import.assert_called_once_with("air.distiller.executor.executor")
        self.assertIs(
            cls,
            mock_module.Executor,
            "get_executor_class should return the Executor class from the mock module.",
        )

    def test_get_executor_class_invalid(self):
        """Tests that get_executor_class raises ValueError for an unknown agent_class."""
        with self.assertRaises(ValueError):
            get_executor_class("UnknownAgent")

    @patch("air.distiller.executor.get_executor_class")
    def test_get_executor_creates_instance(self, mock_get_executor_class):
        """Tests that get_executor creates an instance of the retrieved executor class."""
        # Prepare a MagicMock class to stand in for the real executor class.
        mock_class = MagicMock()
        mock_get_executor_class.return_value = mock_class

        # Call get_executor, which should instantiate the mock class.
        instance = get_executor("CustomAgent", arg1="value1")

        # Verify that get_executor_class was called with the correct agent name.
        mock_get_executor_class.assert_called_once_with("CustomAgent")
        # Verify that the mock class was called (instantiated) with the given arguments.
        mock_class.assert_called_once_with(arg1="value1")
        # The return value should be the new instance.
        self.assertEqual(
            instance,
            mock_class(),
            "get_executor should return the newly instantiated executor object.",
        )


if __name__ == "__main__":
    unittest.main()
