import asyncio
import pytest
from unittest.mock import patch, MagicMock

from air.distiller.executor.google_executor import GoogleExecutor


@pytest.fixture
def send_queue():
    """Fixture providing an asyncio queue for tests."""
    return asyncio.Queue()


@pytest.fixture
def mock_agent_engine():
    """Fixture that returns a mocked Vertex AI AgentEngine."""
    mock_engine = MagicMock()
    # Mock the output of agent_engine.run()
    mock_engine.run.return_value = [
        {"parts": [{"text": "Hello "}, {"text": "from Vertex AI!"}]}
    ]
    return mock_engine


def test_init_success(send_queue, mock_agent_engine):
    """Test successful initialization of GoogleExecutor."""
    utility_config = {"resource_name": "test_resource"}
    with patch(
        "vertexai.agent_engines.get", return_value=mock_agent_engine
    ) as mock_get:
        executor = GoogleExecutor(
            func={},  # Unused in this example
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )

        # Verify agent_engine and interplay
        assert executor.agent_engine is mock_agent_engine
        mock_get.assert_called_once_with("test_resource")


def test_init_missing_resource_name(send_queue):
    """Test that initialization fails if resource_name is missing."""
    utility_config = {}
    with pytest.raises(ValueError) as exc_info:
        GoogleExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert "Missing 'resource_name' in utility_config" in str(exc_info.value)


def test_execute_agent_no_prompt(send_queue, mock_agent_engine):
    """Test that _execute_agent raises ValueError when no prompt is provided."""
    utility_config = {"resource_name": "test_resource"}
    with patch("vertexai.agent_engines.get", return_value=mock_agent_engine):
        executor = GoogleExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )

        with pytest.raises(ValueError) as exc_info:
            executor._execute_agent()
        assert "Missing 'prompt' parameter" in str(exc_info.value)


def test_execute_agent_success(send_queue, mock_agent_engine):
    """Test that _execute_agent successfully returns the agent's string response."""
    utility_config = {"resource_name": "test_resource"}
    with patch("vertexai.agent_engines.get", return_value=mock_agent_engine):
        executor = GoogleExecutor(
            func={},
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )

        result = executor._execute_agent(
            prompt="Hello Agent", session_id="test_session"
        )
        assert result == "Hello from Vertex AI!"
        # Verify the agent engine run method was called with the right parameters
        mock_agent_engine.run.assert_called_once()
        args, kwargs = mock_agent_engine.run.call_args
        assert "session_id" in kwargs
        assert kwargs["session_id"] == "test_session"
