# pylint: disable=redefined-outer-name

import asyncio
import json
from unittest.mock import AsyncMock, MagicMock, patch
import pytest

from air.distiller.executor.mcp_executor import MCPExecutor


@pytest.fixture
def base_kwargs():
    return {
        "func": {},
        "send_queue": asyncio.Queue(),
        "account": "test_account",
        "project": "test_project",
        "uuid": "test_uuid",
        "role": "test_role",
        "utility_config": {"mcp_sse_url": "http://fake-url"},
        "return_string": True,
    }


def test_init_success(base_kwargs):
    """
    Test the successful initialization of MCPExecutor when mcp_sse_url is provided.
    """
    executor = MCPExecutor(**base_kwargs)
    assert executor._sse_url == "http://fake-url"


def test_init_missing_sse_url(base_kwargs):
    """
    Raises a ValueError if mcp_sse_url is missing from config.
    """
    base_kwargs["utility_config"] = {}
    with pytest.raises(ValueError, match="'mcp_sse_url' missing from utility_config"):
        MCPExecutor(**base_kwargs)


@pytest.mark.asyncio
async def test_json_tools_success(base_kwargs):
    """
    Testing the the correct behavior of _json_tools when list_tools() returns valid data.
    """
    executor = MCPExecutor(**base_kwargs)

    mock_tool = MagicMock()
    mock_tool.name = "tool1"
    mock_tool.description = "A test tool"
    mock_tool.inputSchema = {"param1": {"type": "string"}}

    mock_session = AsyncMock()
    mock_session.list_tools.return_value.tools = [mock_tool]

    with patch(
        "air.distiller.executor.mcp_executor._session_context", return_value=AsyncMock()
    ) as mock_context:
        mock_context.return_value.__aenter__.return_value = mock_session
        tools_json = await executor._json_tools()
        tools = json.loads(tools_json)
        assert tools[0]["function"]["name"] == "tool1"


@pytest.mark.asyncio
async def test_json_tools_error(base_kwargs):
    """
    Ensures _json_tools raises an error if session setup fails.
    """
    executor = MCPExecutor(**base_kwargs)
    with patch(
        "air.distiller.executor.mcp_executor._session_context",
        side_effect=Exception("Session setup failed"),
    ):
        with pytest.raises(Exception, match="Session setup failed"):
            await executor._json_tools()


@pytest.mark.asyncio
async def test_invoke_tool_success(base_kwargs):
    """
    Testing correct parsing of tool outputs
    """
    executor = MCPExecutor(**base_kwargs)

    mock_result = MagicMock()
    mock_result.content = [
        MagicMock(text="tool_text", json=None),
        MagicMock(text=None, json={"key": "val"}),
    ]

    mock_session = AsyncMock()
    mock_session.call_tool.return_value = mock_result

    with patch(
        "air.distiller.executor.mcp_executor._session_context", return_value=AsyncMock()
    ) as mock_context:
        mock_context.return_value.__aenter__.return_value = mock_session
        result = await executor._invoke_tool("test_tool", {"a": 1})
        assert "tool_text" in result
        assert '"key": "val"' in result


@pytest.mark.asyncio
async def test_invoke_tool_error(base_kwargs):
    """
    Ensuring exception is raised during tool call when session fails
    """
    executor = MCPExecutor(**base_kwargs)
    with patch(
        "air.distiller.executor.mcp_executor._session_context",
        side_effect=Exception("Bad tool call"),
    ):
        with pytest.raises(Exception, match="Bad tool call"):
            await executor._invoke_tool("tool", {})


@pytest.mark.asyncio
async def test_call_action_list_tools(base_kwargs):
    """
    Testing the success of the _call_ method with action being list_tools
    """
    executor = MCPExecutor(**base_kwargs)

    with patch.object(
        executor, "_json_tools", return_value=json.dumps([{"test": "tool"}])
    ):
        await executor("__reqid__", action="list_tools")
        item = await base_kwargs["send_queue"].get()
        assert item["request_args"]["content"] == json.dumps([{"test": "tool"}])


@pytest.mark.asyncio
async def test_call_action_call_tool(base_kwargs):
    """
    Testing the success of the _call_ method with action being call_tool
    """
    executor = MCPExecutor(**base_kwargs)

    with patch.object(executor, "_invoke_tool", return_value="tool_output"):
        await executor(
            "__reqid__", action="call_tool", tool_name="test_tool", arguments={"x": 1}
        )
        item = await base_kwargs["send_queue"].get()
        assert "tool_output" in item["request_args"]["content"]


@pytest.mark.asyncio
async def test_call_invalid_action(base_kwargs):
    """
    Ensuring invalid actions are not accepted by _call_ method
    """
    executor = MCPExecutor(**base_kwargs)

    await executor("__reqid__", action="invalid_action")
    item = await base_kwargs["send_queue"].get()
    assert "Unknown action" in item["request_args"]["content"]


@pytest.mark.asyncio
async def test_call_tool_missing_name(base_kwargs):
    """
    Ensuring calls with missing tool names are not accepted by _call_ method
    """
    executor = MCPExecutor(**base_kwargs)

    await executor("__reqid__", action="call_tool")  # No tool_name
    item = await base_kwargs["send_queue"].get()
    assert "'tool_name' missing for call_tool action" in item["request_args"]["content"]
