import asyncio
import os
from unittest.mock import patch, MagicMock
import pytest
from air.distiller.executor.salesforce_executor import (
    SalesforceSessionManager,
    SalesforceExecutor,
)

# Sample constants
TEST_DOMAIN = "test.salesforce.com"
TEST_AGENT_ID = "agent123"
TEST_TOKEN = "fake_token"
TEST_SESSION_ID = "session123"
TEST_MESSAGE_TEXT = "Hello"
TEST_RESPONSE_MESSAGE = {"messages": [{"message": "Hi there!"}]}
TEST_UTILITY_CONFIG = {"domain": TEST_DOMAIN, "agent_id": TEST_AGENT_ID}


# --- SalesforceSessionManager Tests ---


@patch("air.distiller.executor.salesforce_executor.requests.post")
@patch.dict(
    os.environ, {"SALESFORCE_CLIENT_KEY": "key", "SALESFORCE_CLIENT_SECRET": "secret"}
)
def test_get_access_token_success(mock_post):
    """
    Test successful token acquisition
    """
    mock_post.return_value.json.return_value = {"access_token": TEST_TOKEN}
    mock_post.return_value.raise_for_status = MagicMock()

    salesforce_client_key = os.environ["SALESFORCE_CLIENT_KEY"]
    salesforce_client_secret = os.environ["SALESFORCE_CLIENT_SECRET"]

    token = SalesforceSessionManager.get_access_token(
        salesforce_client_key, salesforce_client_secret, TEST_DOMAIN
    )
    assert token == TEST_TOKEN
    mock_post.assert_called_once()


@patch("air.distiller.executor.salesforce_executor.requests.post")
@patch.dict(
    os.environ, {"SALESFORCE_CLIENT_KEY": "key", "SALESFORCE_CLIENT_SECRET": "secret"}
)
def test_get_access_token_failure(mock_post):
    """
    Test failing token acquisition
    """
    salesforce_client_key = os.environ["SALESFORCE_CLIENT_KEY"]
    salesforce_client_secret = os.environ["SALESFORCE_CLIENT_SECRET"]

    mock_post.return_value.raise_for_status.side_effect = Exception("HTTP Error")
    with pytest.raises(Exception):
        SalesforceSessionManager.get_access_token(
            salesforce_client_key, salesforce_client_secret, TEST_DOMAIN
        )


@patch("air.distiller.executor.salesforce_executor.requests.post")
def test_initiate_session_success(mock_post):
    """
    Test successful session initiation
    """
    mock_post.return_value.json.return_value = {
        "sessionId": TEST_SESSION_ID,
        "messages": [{"message": "Welcome"}],
    }
    mock_post.return_value.raise_for_status = MagicMock()

    sm = SalesforceSessionManager(TEST_DOMAIN, TEST_AGENT_ID, TEST_TOKEN)
    session_id, sequence_id, welcome = sm.initiate_session()

    assert session_id == TEST_SESSION_ID
    assert sequence_id == 0
    assert welcome == "Welcome"


@patch("air.distiller.executor.salesforce_executor.requests.post")
def test_initiate_session_failure(mock_post):
    """
    Test failing session initiation
    """
    mock_post.return_value.raise_for_status.side_effect = Exception("Bad request")

    sm = SalesforceSessionManager(TEST_DOMAIN, TEST_AGENT_ID, TEST_TOKEN)
    with pytest.raises(Exception):
        sm.initiate_session()


@patch("air.distiller.executor.salesforce_executor.requests.post")
def test_send_message_success(mock_post):
    """
    Test successful message posting
    """
    mock_post.return_value.json.return_value = TEST_RESPONSE_MESSAGE
    mock_post.return_value.raise_for_status = MagicMock()

    sm = SalesforceSessionManager(TEST_DOMAIN, TEST_AGENT_ID, TEST_TOKEN)
    result = sm.send_message(TEST_SESSION_ID, 0, TEST_MESSAGE_TEXT)

    assert result == TEST_RESPONSE_MESSAGE


@patch("air.distiller.executor.salesforce_executor.requests.post")
def test_send_message_failure(mock_post):
    """
    Test failing message posting
    """
    mock_post.return_value.raise_for_status.side_effect = Exception("Failed")

    sm = SalesforceSessionManager(TEST_DOMAIN, TEST_AGENT_ID, TEST_TOKEN)
    with pytest.raises(Exception):
        sm.send_message(TEST_SESSION_ID, 0, TEST_MESSAGE_TEXT)


# --- SalesforceExecutor Tests ---


@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.get_access_token",
    return_value=TEST_TOKEN,
)
@patch("air.distiller.executor.salesforce_executor.SalesforceSessionManager")
def test_executor_init_success(mock_session_mgr_class, _):
    """
    Test successful executor initiation
    """
    mock_session_mgr_instance = mock_session_mgr_class.return_value
    queue = asyncio.Queue()
    executor = SalesforceExecutor(
        func={},
        send_queue=queue,
        account="acc",
        project="proj",
        uuid="uuid",
        role="role",
        utility_config=TEST_UTILITY_CONFIG,
    )
    assert executor.session_manager is mock_session_mgr_instance


def test_executor_init_missing_domain():
    """
    Test missing org domain failure
    """
    config = {"agent_id": "id"}
    queue = asyncio.Queue()
    with pytest.raises(ValueError, match="Missing 'domain'"):
        SalesforceExecutor({}, queue, "acc", "proj", "uuid", "role", config)


def test_executor_init_missing_agent_id():
    """
    Test missing agent ID failure
    """
    config = {"domain": "domain"}
    queue = asyncio.Queue()
    with pytest.raises(ValueError, match="Missing 'agent_id'"):
        SalesforceExecutor({}, queue, "acc", "proj", "uuid", "role", config)


@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.get_access_token",
    side_effect=Exception("Token error"),
)
def test_executor_init_token_failure(_):
    """
    Test failing executor initiation with token error
    """
    config = TEST_UTILITY_CONFIG
    queue = asyncio.Queue()
    with pytest.raises(Exception, match="Token error"):
        SalesforceExecutor({}, queue, "acc", "proj", "uuid", "role", config)


@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.initiate_session",
    return_value=("sess", 0, "Welcome"),
)
@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.send_message",
    return_value=TEST_RESPONSE_MESSAGE,
)
@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.get_access_token",
    return_value=TEST_TOKEN,
)
def test_execute_agent_success(_, __, ___):
    """
    Test successful agent execution
    """
    queue = asyncio.Queue()
    executor = SalesforceExecutor(
        {}, queue, "acc", "proj", "uuid", "role", TEST_UTILITY_CONFIG
    )

    result = executor._execute_agent(prompt="Hi")
    assert result == "Hi there!"


@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.get_access_token",
    return_value=TEST_TOKEN,
)
@patch("air.distiller.executor.salesforce_executor.SalesforceSessionManager")
def test_execute_agent_missing_prompt(
    mock_session_mgr_class, _
):  # pylint: disable=unused-argument
    """
    Test failing agent execution missing prompt
    """
    queue = asyncio.Queue()
    executor = SalesforceExecutor(
        {}, queue, "acc", "proj", "uuid", "role", TEST_UTILITY_CONFIG
    )
    with pytest.raises(ValueError, match="Missing 'prompt'"):
        executor._execute_agent()


@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.initiate_session",
    side_effect=Exception("Session error"),
)
@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.get_access_token",
    return_value=TEST_TOKEN,
)
def test_process_agent_session_failure(
    mock_session_mgr_class, *_
):  # pylint: disable=unused-argument
    """
    Test failing agent execution with session error
    """
    queue = asyncio.Queue()
    executor = SalesforceExecutor(
        {}, queue, "acc", "proj", "uuid", "role", TEST_UTILITY_CONFIG
    )

    with pytest.raises(Exception, match="Session error"):
        asyncio.run(executor._process_agent_request("Hello"))


@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.initiate_session",
    return_value=("sess", 0, "Welcome"),
)
@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.send_message",
    side_effect=Exception("Post failed"),
)
@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.get_access_token",
    return_value=TEST_TOKEN,
)
def test_process_agent_post_failure(
    mock_session_mgr_class, *_
):  # pylint: disable=unused-argument
    """
    Test failing agent execution with message posting error
    """
    queue = asyncio.Queue()
    executor = SalesforceExecutor(
        {}, queue, "acc", "proj", "uuid", "role", TEST_UTILITY_CONFIG
    )

    with pytest.raises(Exception, match="Post failed"):
        asyncio.run(executor._process_agent_request("Hello"))


@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.initiate_session",
    return_value=("sess", 0, "Welcome"),
)
@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.send_message",
    return_value={},
)
@patch(
    "air.distiller.executor.salesforce_executor.SalesforceSessionManager.get_access_token",
    return_value=TEST_TOKEN,
)
@patch("air.distiller.executor.salesforce_executor.SalesforceSessionManager")
def test_process_agent_response_parsing_failure(
    mock_session_mgr_class, *_
):  # pylint: disable=unused-argument
    """
    Test failing agent execution with agent response parsing error
    """
    queue = asyncio.Queue()
    executor = SalesforceExecutor(
        {}, queue, "acc", "proj", "uuid", "role", TEST_UTILITY_CONFIG
    )

    with pytest.raises(Exception):
        asyncio.run(executor._process_agent_request("Hello"))
