# test_tool_executor.py
# pylint: disable=unused-import, too-many-positional-arguments, unused-argument, duplicate-code
import asyncio
import json

import pytest

from air.distiller.executor.tool_executor import ToolExecutor


# Mock functions to be used as tools
def mock_tool_func1(param1, param2):
    return f"Result from tool_func1 with {param1} and {param2}"


async def mock_tool_func2(param1):
    await asyncio.sleep(0.1)
    return f"Result from tool_func2 with {param1}"


def mock_tool_func3():
    return "Result from tool_func3 with no params"


def mock_tool_func4(param1: int, param2: float, param3: str):
    """
    A mock tool function with type hints for param1 (int) and param2 (float),
    and string for param3. Used to test automatic type conversion features.
    """
    return (
        f"param1: {param1} (type {type(param1).__name__}), "
        f"param2: {param2} (type {type(param2).__name__}), "
        f"param3: {param3} (type {type(param3).__name__})"
    )


@pytest.mark.asyncio
async def test_tool_executor_initialization():
    """
    Test that ToolExecutor initializes correctly with valid utility_config.
    """
    send_queue = asyncio.Queue()

    # Mock utility_config with custom tools
    utility_config = {
        "custom_tools": [
            json.dumps({"function": {"name": "tool_func1"}}),
            json.dumps({"function": {"name": "tool_func2"}}),
        ]
    }

    # Functions dictionary
    func_dict = {
        "tool_func1": mock_tool_func1,
        "tool_func2": mock_tool_func2,
        "tool_func3": mock_tool_func3,  # Not included in utility_config
    }

    tool_executor = ToolExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    # Check that tool_executor.func contains only the functions specified in custom_tools
    assert "tool_func1" in tool_executor.func
    assert "tool_func2" in tool_executor.func
    assert "tool_func3" not in tool_executor.func


@pytest.mark.asyncio
async def test_tool_executor_call_with_valid_executor():
    """
    Test that ToolExecutor calls the correct function based on '__executor__'.
    """
    send_queue = asyncio.Queue()

    # Mock utility_config with custom tools
    utility_config = {
        "custom_tools": [
            json.dumps({"function": {"name": "tool_func1"}}),
            json.dumps({"function": {"name": "tool_func2"}}),
        ]
    }

    # Functions dictionary
    func_dict = {
        "tool_func1": mock_tool_func1,
        "tool_func2": mock_tool_func2,
    }

    tool_executor = ToolExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    # Test calling 'tool_func1'
    result = await tool_executor(
        request_id="test_id_1",
        param1="value1",
        param2="value2",
        __executor__="tool_func1",
    )
    assert result == "Result from tool_func1 with value1 and value2"

    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == result

    # Test calling 'tool_func2' (async function)
    result = await tool_executor(
        request_id="test_id_2",
        param1="value1",
        __executor__="tool_func2",
    )
    assert result == "Result from tool_func2 with value1"

    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == result


@pytest.mark.asyncio
async def test_tool_executor_call_with_missing_executor():
    """
    Test that ToolExecutor raises ValueError when '__executor__' is missing.
    """
    send_queue = asyncio.Queue()

    # Mock utility_config with custom tools
    utility_config = {"custom_tools": []}

    func_dict = {}

    tool_executor = ToolExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
    )

    with pytest.raises(ValueError) as excinfo:
        await tool_executor(
            request_id="test_id",
            param1="value1",
            param2="value2",
            # '__executor__' is missing
        )
    # Adjust the assertion to check for the substring
    assert "'__executor__' must be specified in kwargs" in str(excinfo.value)


@pytest.mark.asyncio
async def test_tool_executor_with_missing_function_key():
    """
    Test that ToolExecutor raises ValueError when a tool definition is missing the 'function' key.
    """
    send_queue = asyncio.Queue()

    # Invalid utility_config (missing 'function' key)
    utility_config = {
        "custom_tools": [
            json.dumps({"name": "tool_func1"}),  # Missing 'function' key
        ]
    }

    func_dict = {
        "tool_func1": mock_tool_func1,
    }

    with pytest.raises(ValueError) as excinfo:
        ToolExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
        )
    assert "Missing key" in str(excinfo.value)


@pytest.mark.asyncio
async def test_tool_executor_with_missing_name_key():
    """
    Test that ToolExecutor raises ValueError when a tool definition is missing the 'name' key.
    """
    send_queue = asyncio.Queue()

    # Invalid utility_config (missing 'name' key)
    utility_config = {
        "custom_tools": [
            json.dumps({"function": {}}),  # Missing 'name' key
        ]
    }

    func_dict = {
        "tool_func1": mock_tool_func1,
    }

    with pytest.raises(ValueError) as excinfo:
        ToolExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
        )
    assert "Missing key" in str(excinfo.value)


@pytest.mark.asyncio
async def test_tool_executor_call_with_invalid_executor():
    """
    Test that ToolExecutor raises ValueError when '__executor__' is invalid.
    """
    send_queue = asyncio.Queue()

    # Mock utility_config with custom tools
    utility_config = {
        "custom_tools": [
            json.dumps({"function": {"name": "tool_func1"}}),
        ]
    }

    # Functions dictionary
    func_dict = {
        "tool_func1": mock_tool_func1,
    }

    tool_executor = ToolExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
    )

    with pytest.raises(ValueError) as excinfo:
        await tool_executor(
            request_id="test_id",
            param1="value1",
            __executor__="non_existent_tool",  # Invalid executor
        )
    assert f"Tool 'non_existent_tool' is not available in func" in str(excinfo.value)


@pytest.mark.asyncio
async def test_tool_executor_with_extra_arguments():
    """
    Test that ToolExecutor ignores extra arguments not accepted by the selected function.
    """
    send_queue = asyncio.Queue()

    # Mock utility_config with custom tools
    utility_config = {
        "custom_tools": [
            json.dumps({"function": {"name": "tool_func1"}}),
        ]
    }

    func_dict = {
        "tool_func1": mock_tool_func1,
    }

    tool_executor = ToolExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
    )

    # Call with extra arguments
    result = await tool_executor(
        request_id="test_id",
        param1="value1",
        param2="value2",
        param3="value3",  # Extra positional argument (will be ignored)
        extra_kwarg1="extra1",  # Extra keyword argument (will be ignored)
        __executor__="tool_func1",
    )
    assert result == "Result from tool_func1 with value1 and value2"

    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == result


@pytest.mark.asyncio
async def test_tool_executor_with_missing_required_arguments():
    """
    Test that ToolExecutor raises an error when required arguments are missing.
    """
    send_queue = asyncio.Queue()

    # Mock utility_config with custom tools
    utility_config = {
        "custom_tools": [
            json.dumps({"function": {"name": "tool_func1"}}),
        ]
    }

    func_dict = {
        "tool_func1": mock_tool_func1,
    }

    tool_executor = ToolExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
    )

    # Call without required arguments
    with pytest.raises(RuntimeError) as excinfo:
        await tool_executor(
            request_id="test_id",
            param1="value1",
            # Missing 'param2'
            __executor__="tool_func1",
        )
    # Check that the cause of the RuntimeError is a TypeError
    assert isinstance(excinfo.value.__cause__, TypeError)
    # Optionally, check the exception message
    assert "missing 1 required positional argument" in str(excinfo.value)


@pytest.mark.asyncio
async def test_tool_executor_initialization_error():
    """
    Test that ToolExecutor raises an exception during initialization if utility_config is invalid.
    """
    send_queue = asyncio.Queue()

    # Invalid utility_config (malformed JSON)
    utility_config = {"custom_tools": ["{invalid_json}"]}

    func_dict = {}

    with pytest.raises(Exception):
        ToolExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
        )


@pytest.mark.asyncio
async def test_tool_executor_with_no_custom_tools():
    """
    Test that ToolExecutor initializes correctly when there are no custom tools.
    """
    send_queue = asyncio.Queue()

    utility_config = {"custom_tools": []}  # No tools specified

    func_dict = {
        "tool_func1": mock_tool_func1,
    }

    # Since no tools are specified, func should be empty
    tool_executor = ToolExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
    )

    assert tool_executor.func == {}


@pytest.mark.asyncio
async def test_tool_executor_with_non_string_executor():
    """
    Test that ToolExecutor raises TypeError when '__executor__' is not a string.
    """
    send_queue = asyncio.Queue()

    # Mock utility_config with custom tools
    utility_config = {
        "custom_tools": [
            json.dumps({"function": {"name": "tool_func1"}}),
        ]
    }

    func_dict = {
        "tool_func1": mock_tool_func1,
    }

    tool_executor = ToolExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
    )

    with pytest.raises(TypeError) as excinfo:
        await tool_executor(
            request_id="test_id",
            param1="value1",
            param2="value2",
            __executor__=12345,  # Non-string executor
        )
    assert "'__executor__' must be a string" in str(excinfo.value)


@pytest.mark.asyncio
async def test_tool_executor_with_type_hint_conversion():
    """
    Test that ToolExecutor automatically converts string arguments to int or float
    based on the function signature type hints.
    """
    send_queue = asyncio.Queue()

    # Add our mock_tool_func4 to the utility_config and func_dict
    utility_config = {
        "custom_tools": [
            json.dumps({"function": {"name": "tool_func4"}}),
        ]
    }

    func_dict = {
        "tool_func4": mock_tool_func4,
    }

    tool_executor = ToolExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
    )

    # Test successful conversion
    result = await tool_executor(
        request_id="test_id_type_conv",
        param1="123",
        param2="45.67",
        param3="foo",
        __executor__="tool_func4",
    )
    assert result == (
        "param1: 123 (type int), "
        "param2: 45.67 (type float), "
        "param3: foo (type str)"
    )

    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == result

    # Test scenario with invalid float value to confirm it remains a string
    result = await tool_executor(
        request_id="test_id_type_conv_invalid",
        param1="999",
        param2="not_a_float",
        param3="bar",
        __executor__="tool_func4",
    )
    # We expect param1 will still convert to int (999), but param2 remains string
    assert result == (
        "param1: 999 (type int), "
        "param2: not_a_float (type str), "
        "param3: bar (type str)"
    )

    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == result
