"""
Unit tests for CustomVectorSearchExecutor of Research Agent
"""

# test_analytics_executor.py
# pylint: disable=unused-import, too-many-positional-arguments, unused-argument, duplicate-code, redefined-outer-name

import asyncio
from typing import Callable
from unittest.mock import patch

import pytest

from air.distiller.executor.vector_search_executor import CustomVectorSearchExecutor


@pytest.mark.asyncio
async def test_aisearch_executor_missing_retriever_class():
    """
    Test initializing CustomVectorSearchExecutor with valid Azure AI Search executor_config.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "retriever_config_list": [
            {
                "retriever_name": "Azure Parks internal database",
                # "retriever_class": "CustomRetriever",
                "description": "Data Retriever for all Park information and statistics",
                "aisearch_config": {},
            },
        ]
    }

    with pytest.raises(ValueError) as excinfo:
        executor = CustomVectorSearchExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert "retriever_config.retriever_class is missing in retriever_config." in str(
        excinfo.value
    )


@pytest.mark.asyncio
async def test_aisearch_executor_missing_retriever_name():
    """
    Test initializing CustomVectorSearchExecutor with valid Azure AI Search executor_config.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "retriever_config_list": [
            {
                # "retriever_name": "Azure Parks internal database",
                "retriever_class": "CustomRetriever",
                "description": "Data Retriever for all Park information and statistics",
                "aisearch_config": {},
            },
        ]
    }

    with pytest.raises(ValueError) as excinfo:
        executor = CustomVectorSearchExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert "retriever_config.retriever_name is missing in retriever_config." in str(
        excinfo.value
    )


@pytest.mark.asyncio
async def test_custom_retriever_executor_exists():
    """
    Test valid CustomRetriever added to func_dict.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "retriever_config_list": [
            {
                "retriever_name": "Parks internal database",
                "retriever_class": "CustomRetriever",
                "description": "Data Retriever for all Park information and statistics",
            },
        ]
    }

    # User-defined retriever executor function
    def custom_retriever(query):
        """User-defined aisearch function."""
        return [
            {"result": f"User-defined aisearch executed query: {query}", "score": 0.9}
        ]

    func_dict = {"Parks internal database": custom_retriever}

    executor = CustomVectorSearchExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    assert "Parks internal database" in executor.func


@pytest.mark.asyncio
async def test_custom_retriever_executor_missing():
    """
    Test valid CustomRetriever missing in func_dict.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        "retriever_config_list": [
            {
                "retriever_name": "Parks internal database",
                "retriever_class": "CustomRetriever",
                "description": "Data Retriever for all Park information and statistics",
            },
        ]
    }

    # User-defined retriever executor function
    def custom_retriever(query):
        """User-defined aisearch function."""
        return [
            {"result": f"User-defined aisearch executed query: {query}", "score": 0.9}
        ]

    func_dict = {}

    expected_error = (
        "No executor for custom retriever Parks internal database found. "
        + "You must create this executor and add it to your "
        + "executor_dict."
    )

    with pytest.raises(ValueError) as excinfo:
        executor = CustomVectorSearchExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert expected_error in str(excinfo.value)


@pytest.mark.asyncio
async def test_aisearch_executor_call_with_valid_executor():
    """
    Test that VectorSearchExecutor can execute the correct function based on __executor__.
    """
    send_queue = asyncio.Queue()

    utility_config = {
        "retriever_config_list": [
            {
                "retriever_name": "Parks internal database",
                "retriever_class": "CustomRetriever",
                "description": "Data Retriever for all Park information and statistics",
            },
        ]
    }

    def mock_executor(query):
        """User-defined aisearch function."""
        return [{"result": f"Mock aisearch executed query: {query}", "score": 0.9}]

    func_dict = {"Parks internal database": mock_executor}
    executor = CustomVectorSearchExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    # Now call the executor with query as a positional argument
    result = await executor(
        "test_id",
        query="test query",
        __executor__="Parks internal database",
    )

    # expected_result = "Executed custom azure ai search with query: test query."
    expected_result = [
        {"result": "Mock aisearch executed query: test query", "score": 0.9}
    ]
    assert result == expected_result

    # Check that send_queue has received the message
    sent_message = await send_queue.get()
    assert sent_message["request_args"]["content"] == expected_result


@pytest.mark.asyncio
async def test_aisearch_executor_call_with_invalid_executor():
    """
    Test that VectorSearchExecutor raises error when using invalid __executor__.
    """
    send_queue = asyncio.Queue()

    utility_config = {
        "retriever_config_list": [
            {
                "retriever_name": "Parks internal database",
                "retriever_class": "CustomRetriever",
                "description": "Data Retriever for all Park information and statistics",
            },
        ]
    }

    def mock_executor(query):
        """User-defined aisearch function."""
        return [{"result": f"Mock aisearch executed query: {query}", "score": 0.9}]

    func_dict = {"Parks internal database": mock_executor}

    executor = CustomVectorSearchExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    # Try calling with invalid __executor__
    with pytest.raises(ValueError) as excinfo:
        await executor(
            request_id="test_id",
            query="test query",
            __executor__="InvalidExecutor",
        )
    assert "Retriever type 'InvalidExecutor' is not available in func." in str(
        excinfo.value
    )


@pytest.mark.asyncio
async def test_aisearch_executor_missing_retriever_config_list():
    """
    Test that CustomVectorSearchExecutor raises error when retriever_config_list is missing.
    """
    send_queue = asyncio.Queue()
    func_dict = {}
    utility_config = {
        # "retriever_config_list": [
        #     {
        #         "retriever_name": "Azure Parks internal database",
        #         "retriever_class": "CustomRetriever",
        #         "description": "Data Retriever for all Park information and statistics",
        #         "aisearch_config": {}
        #     },
        # ]
    }

    with pytest.raises(ValueError) as excinfo:
        CustomVectorSearchExecutor(
            func=func_dict,
            send_queue=send_queue,
            account="test_account",
            project="test_project",
            uuid="test_uuid",
            role="test_role",
            utility_config=utility_config,
            return_string=True,
        )
    assert "retriever_config_list is missing in utility_config." in str(excinfo.value)


@pytest.mark.asyncio
async def test_aisearch_executor_missing_executor_in_kwargs():
    """
    Test that __executor__ is provided in kwargs
    """
    send_queue = asyncio.Queue()

    # User-defined retriever executor function
    def mock_executor(query):
        """User-defined aisearch function."""
        return [{"result": f"Mock aisearch executed query: {query}", "score": 0.9}]

    func_dict = {"Parks internal database": mock_executor}

    utility_config = {
        "retriever_config_list": [
            {
                "retriever_name": "Parks internal database",
                "retriever_class": "CustomRetriever",
                "description": "Data Retriever for all Park information and statistics",
                "aisearch_config": {"top_k": 5},
            },
        ]
    }

    executor = CustomVectorSearchExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    # Ensure that the user-defined function is used

    with pytest.raises(ValueError) as excinfo:
        result = await executor(
            request_id="test_id",
            query="test query",
            # __executor__="User-defined aisearch function",
        )
    assert "'__executor__' must be specified in kwargs." in str(excinfo.value)


@pytest.mark.asyncio
async def test_validate_result_with_valid_data():
    """
    Unit test for validating results with valid data
    """
    send_queue = asyncio.Queue()
    func_dict = {
        "Parks internal database": lambda query: [
            {"result": "Mock result", "score": 1.0}
        ]
    }
    utility_config = {
        "retriever_config_list": [
            {
                "retriever_name": "Parks internal database",
                "retriever_class": "CustomRetriever",
                "description": "Data Retriever for all Park information and statistics",
            }
        ]
    }

    executor = CustomVectorSearchExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    valid_result = [
        {"result": "some result", "score": 1.0},
        {"result": "another result", "score": 0.8},
    ]

    assert executor.validate_result(valid_result) == valid_result


@pytest.mark.asyncio
async def test_validate_result_with_empty_list():
    """
    Unit test for validating results when result is empty list
    """
    send_queue = asyncio.Queue()
    func_dict = {
        "Parks internal database": lambda query: [
            {"result": "Mock result", "score": 1.0}
        ]
    }
    utility_config = {
        "retriever_config_list": [
            {
                "retriever_name": "Parks internal database",
                "retriever_class": "CustomRetriever",
                "description": "Data Retriever for all Park information and statistics",
            }
        ]
    }

    executor = CustomVectorSearchExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    assert executor.validate_result([]) == []


@pytest.mark.asyncio
async def test_validate_result_with_invalid_data():
    send_queue = asyncio.Queue()
    func_dict = {
        "Parks internal database": lambda query: [
            {"result": "Mock result", "score": 1.0}
        ]
    }
    utility_config = {
        "retriever_config_list": [
            {
                "retriever_name": "Parks internal database",
                "retriever_class": "CustomRetriever",
                "description": "Data Retriever for all Park information and statistics",
            }
        ]
    }

    executor = CustomVectorSearchExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    invalid_result = [
        {"result": "some result"},  # Missing 'score'
        {"score": 0.8},  # Missing 'result'
        {"result": 123, "score": 0.9},  # 'result' is not a string
        {"result": "valid result", "score": "high"},  # 'score' is not numeric
    ]

    assert executor.validate_result(invalid_result) == []


@pytest.mark.asyncio
async def test_validate_result_with_non_list_input():
    send_queue = asyncio.Queue()
    func_dict = {
        "Parks internal database": lambda query: [
            {"result": "Mock result", "score": 1.0}
        ]
    }
    utility_config = {
        "retriever_config_list": [
            {
                "retriever_name": "Parks internal database",
                "retriever_class": "CustomRetriever",
                "description": "Data Retriever for all Park information and statistics",
            }
        ]
    }

    executor = CustomVectorSearchExecutor(
        func=func_dict,
        send_queue=send_queue,
        account="test_account",
        project="test_project",
        uuid="test_uuid",
        role="test_role",
        utility_config=utility_config,
        return_string=True,
    )

    non_list_input = {"result": "some result", "score": 1.0}

    assert executor.validate_result(non_list_input) == []
