import asyncio

# pylint: disable=protected-access, unused-argument, f-string-without-interpolation
import os
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from httpx import Response
from requests.exceptions import RequestException

from air.distiller.executor.writer_executor import WriterExecutor


@pytest.fixture
def send_queue():
    """Fixture providing an asyncio queue for tests."""
    return asyncio.Queue()


@pytest.fixture(autouse=True)
def clear_env(monkeypatch):
    """Ensure our tests control the env var entirely."""
    monkeypatch.delenv("MY_ENV_WRITER_KEY", raising=False)


@pytest.fixture
def utility_config(monkeypatch):
    """Provide a valid utility_config and set the corresponding env var."""
    monkeypatch.setenv("MY_ENV_WRITER_KEY", "secret_value")
    return {
        "application_id": "app123",  # app identifier for Writer AI
        "api_key_env_var": "MY_ENV_WRITER_KEY",
    }


def test_init_success(send_queue, utility_config):
    """Test successful initialization with required config."""
    executor = WriterExecutor(
        func={},
        send_queue=send_queue,
        account="acct",
        project="proj",
        uuid="uuid",
        role="role",
        utility_config=utility_config,
        return_string=True,
    )
    assert executor.application_id == "app123"
    assert executor.writer_key == "secret_value"


def test_init_missing_application_id(send_queue, monkeypatch):
    """Test init raises if 'application_id' missing from config."""
    monkeypatch.setenv("MY_ENV_WRITER_KEY", "secret_value")
    with pytest.raises(ValueError) as exc:
        WriterExecutor(
            func={},
            send_queue=send_queue,
            account="a",
            project="p",
            uuid="u",
            role="r",
            utility_config={"api_key_env_var": "MY_ENV_WRITER_KEY"},
            return_string=True,
        )
    assert "Missing required configuration 'application_id_str'" in str(exc.value)


def test_init_missing_api_key_env_var(send_queue):
    """Test init raises if 'api_key_env_var' missing from config."""
    with pytest.raises(ValueError) as exc:
        WriterExecutor(
            func={},
            send_queue=send_queue,
            account="a",
            project="p",
            uuid="u",
            role="r",
            utility_config={"application_id": "app123"},
            return_string=True,
        )
    assert "Missing required configuration 'api_key_env_var'" in str(exc.value)


def test_init_missing_env_var(send_queue, utility_config):
    """Test init raises if the named env var is not set."""
    # delete it to simulate not present
    os.environ.pop("MY_ENV_WRITER_KEY", None)
    with pytest.raises(ValueError) as exc:
        WriterExecutor(
            func={},
            send_queue=send_queue,
            account="a",
            project="p",
            uuid="u",
            role="r",
            utility_config=utility_config,
            return_string=True,
        )
    assert "Environment variable 'MY_ENV_WRITER_KEY' is not set or is empty." in str(
        exc.value
    )


def test_execute_agent_missing_action(send_queue, utility_config):
    """Test that missing 'action' raises ValueError."""
    executor = WriterExecutor(
        func={},
        send_queue=send_queue,
        account="a",
        project="p",
        uuid="u",
        role="r",
        utility_config=utility_config,
    )
    with pytest.raises(ValueError, match="Missing 'action' parameter"):
        executor._execute_agent()


def test_execute_agent_unknown_action(send_queue, utility_config):
    """Test that unknown 'action' raises ValueError."""
    executor = WriterExecutor(
        func={},
        send_queue=send_queue,
        account="a",
        project="p",
        uuid="u",
        role="r",
        utility_config=utility_config,
    )
    with pytest.raises(ValueError, match="Unknown action 'foo'"):
        executor._execute_agent(action="foo")


def test_execute_agent_request_input_schema_dispatch(send_queue, utility_config):
    """Test dispatch for 'request_input_schema'."""
    executor = WriterExecutor(
        func={},
        send_queue=send_queue,
        account="a",
        project="p",
        uuid="u",
        role="r",
        utility_config=utility_config,
    )
    fake_schema = '[{"inputs": []}]'
    executor._request_input_schema = AsyncMock(return_value=fake_schema)
    with patch(
        "air.distiller.executor.writer_executor.asyncio.run", return_value=fake_schema
    ) as run_mock:
        result = executor._execute_agent(action="request_input_schema")
        run_mock.assert_called_once()
        assert result == fake_schema


def test_execute_agent_process_inputs_dispatch(send_queue, utility_config):
    """Test dispatch for 'process_inputs'."""
    executor = WriterExecutor(
        func={},
        send_queue=send_queue,
        account="a",
        project="p",
        uuid="u",
        role="r",
        utility_config=utility_config,
    )
    fake_resp = "RESULT"
    executor._process_inputs = AsyncMock(return_value=fake_resp)
    with patch(
        "air.distiller.executor.writer_executor.asyncio.run", return_value=fake_resp
    ) as run_mock:
        result = executor._execute_agent(action="process_inputs", inputs=[1, 2, 3])
        run_mock.assert_called_once()
        assert result == fake_resp


@pytest.mark.asyncio
async def test_request_input_schema_success(send_queue, utility_config):
    """Test _request_input_schema returns inputs field and calls AsyncWriter.get."""
    executor = WriterExecutor(
        func={},
        send_queue=send_queue,
        account="a",
        project="p",
        uuid="u",
        role="r",
        utility_config=utility_config,
    )
    executor.writer_key = "KEY"
    fake_resp = MagicMock()
    fake_resp.json.return_value = {"inputs": ["a"]}
    fake_client = MagicMock()
    fake_client.get = AsyncMock(return_value=fake_resp)
    with patch(
        "air.distiller.executor.writer_executor.AsyncWriter", return_value=fake_client
    ):
        schema = await executor._request_input_schema()
        assert schema == ["a"]
        fake_client.get.assert_awaited_once_with(
            "/v1/applications/app123", cast_to=Response
        )


@pytest.mark.asyncio
async def test_request_input_schema_failure(send_queue, utility_config):
    """Test _request_input_schema raises on RequestException."""
    executor = WriterExecutor(
        func={},
        send_queue=send_queue,
        account="a",
        project="p",
        uuid="u",
        role="r",
        utility_config=utility_config,
    )
    fake_client = MagicMock()
    fake_client.get = AsyncMock(side_effect=RequestException("oops"))
    with patch(
        "air.distiller.executor.writer_executor.AsyncWriter", return_value=fake_client
    ):
        with pytest.raises(RuntimeError, match="Failed to fetch input schema"):
            await executor._request_input_schema()


@pytest.mark.asyncio
async def test_process_inputs_no_job_id(send_queue, utility_config):
    """Test _process_inputs wraps missing job ID in RuntimeError."""
    executor = WriterExecutor(
        func={},
        send_queue=send_queue,
        account="a",
        project="p",
        uuid="u",
        role="r",
        utility_config=utility_config,
    )
    executor.writer_key = "KEY"
    fake_post_resp = MagicMock()
    fake_post_resp.json.return_value = {}
    fake_client = MagicMock()
    fake_client.post = AsyncMock(return_value=fake_post_resp)
    with patch(
        "air.distiller.executor.writer_executor.AsyncWriter", return_value=fake_client
    ):
        with pytest.raises(
            RuntimeError,
            match="Exception occurred while submitting job: Failed to start job",
        ):
            await executor._process_inputs([{}])


@pytest.mark.asyncio
async def test_process_inputs_job_failed(send_queue, utility_config):
    """Test _process_inputs wraps job failure status in RuntimeError."""
    executor = WriterExecutor(
        func={},
        send_queue=send_queue,
        account="a",
        project="p",
        uuid="u",
        role="r",
        utility_config=utility_config,
    )
    executor.writer_key = "KEY"
    fake_post_resp = MagicMock()
    fake_post_resp.json.return_value = {"id": "job1"}
    fake_get_resp = MagicMock()
    fake_get_resp.json.return_value = {"status": "failed", "data": {}}
    fake_client = MagicMock()
    fake_client.post = AsyncMock(return_value=fake_post_resp)
    fake_client.get = AsyncMock(return_value=fake_get_resp)
    with patch(
        "air.distiller.executor.writer_executor.AsyncWriter", return_value=fake_client
    ), patch.object(asyncio, "sleep", AsyncMock()):
        with pytest.raises(
            RuntimeError,
            match="Exception occurred while polling job status: Job failed",
        ):
            await executor._process_inputs([{}])


@pytest.mark.asyncio
async def test_process_inputs_success_flow(send_queue, utility_config):
    """Test _process_inputs returns raw suggestion string on completion."""
    executor = WriterExecutor(
        func={},
        send_queue=send_queue,
        account="a",
        project="p",
        uuid="u",
        role="r",
        utility_config=utility_config,
    )
    executor.writer_key = "KEY"
    fake_post_resp = MagicMock()
    fake_post_resp.json.return_value = {"id": "job1"}
    seq = [
        {"status": "in_progress"},
        {"status": "completed", "data": {"suggestion": '{"foo": "bar"}'}},
    ]

    async def fake_get(*args, **_kwargs):
        m = MagicMock()
        m.json.return_value = seq.pop(0)
        return m

    fake_client = MagicMock()
    fake_client.post = AsyncMock(return_value=fake_post_resp)
    fake_client.get = AsyncMock(side_effect=fake_get)
    with patch(
        "air.distiller.executor.writer_executor.AsyncWriter", return_value=fake_client
    ), patch.object(asyncio, "sleep", AsyncMock()):
        result = await executor._process_inputs([{}])
        assert result == '{"foo": "bar"}'  # raw suggestion string
