import gc  # Add garbage collection
import os
from unittest.mock import patch

import pytest
from presidio_analyzer import RecognizerResult

from air.distiller.pii_handler.pii_handler import PIIHandler, PIIMetadata


@pytest.fixture(scope="module")  # Changed to module scope to reuse
def pii_handler():
    """Fixture to initialize PIIHandler with config."""
    config_path = os.path.join(
        os.path.dirname(__file__),
        "../../../../air/distiller/pii_handler/pii_handler.yaml",
    )
    handler = PIIHandler(config_path=config_path)
    handler.enable()
    yield handler
    # Force cleanup after module tests are complete
    handler.clear_mapping()
    del handler.analyzer
    del handler.anonymizer
    del handler.deanonymizer
    gc.collect()  # Explicitly request garbage collection


@pytest.fixture(autouse=True)
def reset_between_tests(pii_handler):
    """Reset state between tests to prevent memory buildup."""
    pii_handler.clear_mapping()
    yield
    pii_handler.clear_mapping()
    gc.collect()  # Explicitly collect garbage after each test


def test_is_enabled(pii_handler):
    """Test enabling and disabling PIIHandler."""
    assert pii_handler.is_enabled() is True
    pii_handler.disable()
    assert pii_handler.is_enabled() is False
    pii_handler.enable()  # Re-enable for subsequent tests


def test_clear_mapping(pii_handler):
    """Test clearing entity_counts and pii_mapping."""
    text = "Hi, my name is Henry."
    _, _ = pii_handler.mask_text(text)
    assert pii_handler.entity_counts
    pii_handler.clear_mapping()
    assert pii_handler.entity_counts == {}
    assert pii_handler.pii_mapping == {}


def test_mask_with_unknown_entity(pii_handler):
    """Test masking when entity is not configured."""
    original = "Sensitive: XYZ123"
    fake_result = [
        RecognizerResult(entity_type="UNCONFIGURED_ENTITY", start=11, end=17, score=0.9)
    ]
    with patch.object(pii_handler.analyzer, "analyze", return_value=fake_result):
        masked, metadata = pii_handler.mask_text(original)
        assert "[UNCONFIGURED_ENTITY_1]" in masked
        assert any(m.entity_type == "UNCONFIGURED_ENTITY" for m in metadata)


def test_demask_with_no_placeholders(pii_handler):
    """Test demasking when no placeholders exist in text."""
    original = "Nothing to demask here."
    demasked = pii_handler.demask_text(original, [])
    assert demasked == original


def test_mask_text_no_entities(pii_handler):
    """Test that masking does nothing if no PII is found."""
    text = "Just some harmless sentence."
    masked, metadata = pii_handler.mask_text(text)
    assert masked == text
    assert metadata == []


def test_demask_with_missing_placeholder_field(pii_handler):
    """Test demask fallback behavior when placeholder field is missing."""
    text = "Name: Henry"
    masked, metadata = pii_handler.mask_text(text)

    # Create modified metadata with non-existent placeholders to simulate missing field
    modified_metadata = []
    for m in metadata:
        # Create a new PIIMetadata object with a placeholder that doesn't exist in text
        # This simulates the scenario where placeholder field is missing/corrupted
        modified_meta = PIIMetadata(
            mapping_key=m.mapping_key,
            original_text=m.original_text,
            entity_type=m.entity_type,
            operator=m.operator,
            start=m.start,
            end=m.end,
            placeholder="[NONEXISTENT_PLACEHOLDER]",  # Placeholder that won't be found
        )
        modified_metadata.append(modified_meta)

    demasked = pii_handler.demask_text(masked, modified_metadata)
    # With non-matching placeholders, text should remain masked
    assert demasked == masked


def test_demask_placeholder_not_found(pii_handler):
    """Test demask does not change text if placeholder not found."""
    masked = "This text has [EMAIL_1] but the metadata is wrong."
    wrong_metadata = [
        PIIMetadata(
            mapping_key="PHONE_0_3",
            original_text="123",
            entity_type="PHONE",
            operator="replace",
            start=0,
            end=3,
            placeholder="[PHONE_1]",
        )
    ]
    demasked = pii_handler.demask_text(masked, wrong_metadata)
    assert demasked == masked


def test_reuse_existing_metadata(pii_handler):
    """Test reuse of placeholder from existing metadata."""
    text1 = "My name is Henry."
    _, metadata1 = pii_handler.mask_text(text1)
    # Emulate memory by explicitly storing prior metadata
    pii_handler.extend_metadata(metadata1)
    text2 = "Henry is here again."
    masked2, metadata2 = pii_handler.mask_text(text2)
    assert "[PERSON_1]" in masked2  # Reuse same placeholder
    assert len(metadata2) == 1


def test_mask_json_with_top_level_list(pii_handler):
    """Test masking when top-level JSON is a list."""
    data = ["Email: henry@example.com"]
    wrapped = {"top": data}
    _, meta = pii_handler.mask_json(wrapped)
    assert any(m.entity_type == "EMAIL_ADDRESS" for m in meta)


def test_demask_json_with_top_level_list(pii_handler):
    """Test demask of JSON with top-level list."""
    data = ["Email: henry@example.com"]
    masked, meta = pii_handler.mask_json({"list": data})
    demasked = pii_handler.demask_json({"list": masked["list"]}, meta)
    assert demasked["list"] == data


def test_demask_json_list_fallback(pii_handler):
    """Test demask when input is a list instead of a dict."""
    masked = ["[EMAIL_ADDRESS_1]"]
    metadata = [
        PIIMetadata(
            mapping_key="EMAIL_ADDRESS_0_19",
            entity_type="EMAIL_ADDRESS",
            original_text="henry@example.com",
            operator="replace",
            start=0,
            end=19,
            placeholder="[EMAIL_ADDRESS_1]",
        )
    ]
    restored = pii_handler.demask_json(masked, metadata)
    assert restored == ["henry@example.com"]


def test_empty_input(pii_handler):
    """Test masking empty input."""
    masked, _ = pii_handler.mask_text("")
    assert masked == ""


def test_disable_pii_masking(pii_handler):
    """Test masking is bypassed when PIIHandler is disabled."""
    pii_handler.disable()
    text = "Hi, I'm Henry. My number is 4111 1111 1111 1111"
    masked, metadata = pii_handler.mask_text(text)
    assert masked == text
    assert metadata == []
    pii_handler.enable()  # Re-enable for subsequent tests


def test_mask_json_disabled(pii_handler):
    """Test masking is skipped for JSON if handler is disabled."""
    pii_handler.disable()
    data = {"email": "henry@example.com"}
    masked, meta = pii_handler.mask_json(data)
    assert masked == data
    assert meta == []
    pii_handler.enable()  # Re-enable for subsequent tests


def test_mask_json(pii_handler):
    """Test that PII in nested JSON is masked properly."""
    data = {
        "user": "Henry",
        "card": "4111 1111 1111 1111",
        "nested": {"email": "henry@example.com"},
    }
    masked_json, metadata = pii_handler.mask_json(data)
    assert "CREDIT_CARD" in str(masked_json)  # Just check types, not exact strings
    assert "EMAIL_ADDRESS" in str(masked_json)
    assert len(metadata) >= 1


def test_demask_json(pii_handler):
    """Test that masked JSON is correctly demasked."""
    original = {
        "user": "Henry",
        "card": "4111 1111 1111 1111",
        "nested": {"email": "henry@example.com"},
    }
    masked, metadata = pii_handler.mask_json(original)
    restored = pii_handler.demask_json(masked, metadata)
    assert restored == original


def test_mask_json_with_mixed_types(pii_handler):
    """Test masking leaves primitive values unmodified."""
    data = {"name": "Henry", "age": 30, "valid": True}
    masked, _ = pii_handler.mask_json(data)
    assert masked["age"] == 30
    assert masked["valid"] is True


def test_demask_json_without_metadata(pii_handler):
    """Test demasking JSON with empty metadata."""
    data = {"note": "Nothing masked here."}
    restored = pii_handler.demask_json(data, [])
    assert restored == data


def test_demask_json_with_primitive_value(pii_handler):
    """Test demasking primitive value in JSON."""
    data = {"value": 42}
    restored = pii_handler.demask_json(data, metadata=[])
    assert restored == {"value": 42}
