# pylint: disable=unused-import, too-many-positional-arguments, unused-argument

import asyncio
import json
import sys
from asyncio import Queue
from unittest.mock import AsyncMock, MagicMock, Mock, patch

import pytest
from omegaconf import OmegaConf

from air.api import PostgresAPI
from air.distiller.client import AsyncDistillerClient, string_check

# ---------------------- FIXTURES ---------------------- #


# Test cases for string_check
def test_valid_string():
    """Tests if valid strings pass the string_check function."""
    valid_strings = ["air", "123", "air-123", "A_I_R", "test123"]
    for s in valid_strings:
        assert string_check(s) is None


def test_invalid_string():
    """Tests if invalid strings raise a ValueError."""
    invalid_strings = ["air!@#", "123+456", "a i r", "test@test", "$special_char"]
    for s in invalid_strings:
        with pytest.raises(ValueError):
            string_check(s)


def test_empty_string():
    """Tests if an empty string raises a ValueError."""
    with pytest.raises(ValueError):
        string_check("")


@pytest.fixture(autouse=True)
def fx_mock_auth():
    """
    Automatically mock 'air.distiller.client.auth' module globally.
    """
    with patch("air.distiller.client.auth") as mocked_auth:
        mocked_auth.account = "test-account"
        mocked_auth.get_access_token.return_value = "test-token"
        yield mocked_auth


@pytest.fixture
def fx_mock_pii_config():
    """Return a mock configuration for PIIHandler."""
    return OmegaConf.create(
        {
            "common_entities": [
                "PERSON",
                "PHONE_NUMBER",
                "EMAIL_ADDRESS",
                "CREDIT_CARD",
                "US_SSN",
                "US_BANK_NUMBER",
                "US_PASSPORT",
                "LOCATION",
                "DATE_TIME",
                "IP_ADDRESS",
            ],
            "entity_operator_mapping": {
                "CREDIT_CARD": {
                    "operator": "replace",
                    "params": {"new_value": "[CREDIT_CARD]"},
                },
                "US_SSN": {"operator": "replace", "params": {"new_value": "[US_SSN]"}},
                "US_BANK_NUMBER": {
                    "operator": "replace",
                    "params": {"new_value": "[US_BANK_NUMBER]"},
                },
                "US_PASSPORT": {
                    "operator": "replace",
                    "params": {"new_value": "[US_PASSPORT]"},
                },
                "PERSON": {"operator": "replace", "params": {"new_value": "[PERSON]"}},
                "PHONE_NUMBER": {
                    "operator": "replace",
                    "params": {"new_value": "[PHONE]"},
                },
                "EMAIL_ADDRESS": {
                    "operator": "replace",
                    "params": {"new_value": "[EMAIL]"},
                },
                "LOCATION": {
                    "operator": "replace",
                    "params": {"new_value": "[LOCATION]"},
                },
                "DATE_TIME": {"operator": "replace", "params": {"new_value": "[DATE]"}},
                "IP_ADDRESS": {"operator": "replace", "params": {"new_value": "[IP]"}},
                "DEFAULT": {"operator": "replace", "params": {"new_value": "<PII>"}},
            },
        }
    )


@pytest.fixture
def fx_distiller_client(fx_mock_pii_config, request):
    """
    Provide a AsyncDistillerClient and guarantee it is closed after every test
    so the background _send_loop / _ping_monitor tasks do not leak.
    """
    with patch("omegaconf.OmegaConf.load", return_value=fx_mock_pii_config):
        client = AsyncDistillerClient()

    def _cleanup() -> None:
        """
        Run client.close() on whichever event loop is still available.
        If pytest has already disposed of the main loop we create a
        temporary one just to finish the shutdown cleanly.
        """
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        if not loop.is_closed():
            loop.run_until_complete(client.close())

    request.addfinalizer(_cleanup)
    return client


@pytest.fixture
def fx_config_data():
    """Return sample config data."""
    return {"model": "gpt-4", "temperature": 0.7, "settings": {"max_tokens": 1000}}


@pytest.fixture
def fx_mock_websocket():
    """Return a mocked async websocket connection."""
    with patch(
        "air.distiller.client.websockets.connect", new_callable=AsyncMock
    ) as mock_ws:
        mock_ws.return_value.send = AsyncMock()
        mock_ws.return_value.recv = AsyncMock()
        mock_ws.return_value.close = AsyncMock()
        yield mock_ws


@pytest.fixture
def fx_mock_db_client():
    """Return a mocked PostgresAPI with async 'execute_query'."""
    mock_db = MagicMock(spec=PostgresAPI)
    mock_db.execute_query = AsyncMock()
    return mock_db


@pytest.fixture
def mock_requests_post(mocker):
    mock_project_config = (
        '"{\\"orchestrator\\": {\\"agent_list\\": ['
        '{\\"agent_name\\": \\"Search Agent\\"}'
        "]}, "
        '\\"utility_agents\\": ['
        '{\\"agent_class\\": \\"SearchAgent\\", '
        '\\"agent_name\\": \\"Search Agent\\"}'
        "]}"
        '"'
    )
    return mocker.patch(
        "requests.post",
        return_value=mocker.Mock(
            status_code=200, text=json.dumps({"config": mock_project_config})
        ),
    )


# ---------------------- TESTS ---------------------- #


def test_create_project_success(request):
    """
    Test 'create_project' succeeds when API returns 201.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    config_data = request.getfixturevalue("fx_config_data")

    mock_response = MagicMock()
    mock_response.status_code = 201
    mock_response.content = json.dumps({"project_version": -1})

    with patch("requests.post", return_value=mock_response) as mock_post, patch(
        "omegaconf.OmegaConf.load", return_value=config_data
    ) as mock_load, patch(
        "omegaconf.OmegaConf.to_container", return_value=config_data
    ) as mock_to_container:
        result = distiller_client.create_project(
            config_path="test_config.yaml", project="test_project"
        )
        assert result is True
        mock_load.assert_called_once_with("test_config.yaml")
        mock_to_container.assert_called_once_with(config_data, resolve=True)
        mock_post.assert_called_once()


def test_create_project_failure(request):
    """
    Test 'create_project' fails when API returns non-201 status code.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    config_data = request.getfixturevalue("fx_config_data")

    mock_response = MagicMock()
    mock_response.status_code = 400

    with patch("requests.post", return_value=mock_response), patch(
        "omegaconf.OmegaConf.load", return_value=config_data
    ), patch("omegaconf.OmegaConf.to_container", return_value=config_data):
        result = distiller_client.create_project(
            config_path="test_config.yaml", project="test_project"
        )
        assert result is False


def test_download_project_success(request):
    """
    Test 'download_project' returns expected config when successful.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    expected_config = {"some": "config"}

    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = json.dumps(expected_config)

    with patch("requests.post", return_value=mock_response):
        result = distiller_client.download_project("test_project")
        assert result == expected_config


def test_download_project_failure(request):
    """
    Test 'download_project' returns None when the API call fails.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_response = MagicMock()
    mock_response.status_code = 400

    with patch("requests.post", return_value=mock_response):
        result = distiller_client.download_project("test_project")
        assert result is None


@pytest.mark.asyncio
async def test_connect_success(request, mock_requests_post):
    """
    Test 'connect' method successfully establishes a websocket connection.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.return_value.recv.side_effect = [
        '{"type": "PING"}',
        '{"type": "data", "value": "test_message"}',
    ]

    await distiller_client.connect(
        project="test_project", uuid="test_uuid", project_version="0"
    )

    # pylint: disable=protected-access
    assert distiller_client.connection is not None
    assert distiller_client._send_task is not None
    assert distiller_client._receive_task is not None
    assert distiller_client._ping_task is not None


@pytest.mark.asyncio
async def test_connect_failure(request, mock_requests_post):
    """
    Test that 'connect' raises an exception if the connection fails.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.side_effect = Exception("Connection failed")
    with pytest.raises(Exception, match="Connection failed"):
        await distiller_client.connect(
            project="test_project", uuid="test_uuid", project_version="0"
        )

    assert distiller_client.connection is None


@pytest.mark.asyncio
async def test_client_send_and_receive(request, mock_requests_post):
    """
    Test that 'send' and 'recv' work properly with the websocket connection.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.return_value.recv.side_effect = [
        '{"type": "PING"}',
        '{"type": "data", "value": "test_message"}',
    ]

    await distiller_client.connect(
        project="test_project", uuid="test_uuid", project_version="0"
    )
    await distiller_client.send({"type": "PING"})
    message = await distiller_client.recv()
    assert message["type"] == "data"
    assert message["value"] == "test_message"


@pytest.mark.skipif(sys.platform == "win32", reason="Not supported on Windows")
@pytest.mark.asyncio
async def test_ping_monitor_timeout(request, mock_requests_post):
    """
    Test that the client closes the connection on ping timeout.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_connection = AsyncMock()
    distiller_client.connection = mock_connection
    distiller_client.close = AsyncMock()
    mock_websocket.return_value.recv = AsyncMock(
        side_effect=['{"type": "PING"}', '{"type": "data", "value": "test_message"}']
    )

    # pylint: disable=protected-access
    await distiller_client.connect(
        project="test_project", uuid="1234", project_version="0"
    )
    distiller_client._last_ping_received = asyncio.get_event_loop().time() - (
        distiller_client.ping_interval * 2 + 1
    )
    await asyncio.sleep(distiller_client.ping_interval * 2 + 1)
    distiller_client.close.assert_called_once()


@pytest.mark.asyncio
async def test_client_close(request, mock_requests_post):
    """
    Test 'close' properly closes an existing websocket connection.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    # pylint: disable=protected-access
    mock_websocket.return_value.recv.side_effect = [
        '{"type": "PING"}',
        '{"type": "data", "value": "test_message"}',
    ]

    await distiller_client.connect(
        project="test_project", uuid="test_uuid", project_version="0"
    )
    await distiller_client.close()

    assert distiller_client.connection is None
    assert distiller_client.project is None
    assert distiller_client.uuid is None
    assert distiller_client._send_task.done() or distiller_client._send_task.cancelled()
    assert (
        distiller_client._receive_task.done()
        or distiller_client._receive_task.cancelled()
    )
    assert distiller_client._ping_task.done() or distiller_client._ping_task.cancelled()


@pytest.mark.asyncio
async def test_request_valid_response(request):
    """
    Test that 'request' successfully receives a valid response over websocket.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.return_value.recv.side_effect = [
        {"status": "wait", "role": "assistant", "query": "test"},
        {"status": "complete"},
    ]
    distiller_client.project = "test_project"
    distiller_client.uuid = "test_uuid"

    distiller_client.send_queue = Queue()
    await distiller_client.send_queue.put({"mock": "payload"})
    distiller_client.receive_queue = Queue()
    await distiller_client.receive_queue.put({"status": "complete"})

    responses = []
    async for response in distiller_client.request("query", {"query": "test_query"}):
        responses.append(response)
    assert len(responses) == 0


@pytest.mark.asyncio
async def test_request_no_project_or_uuid(request):
    """
    Test that 'request' handles scenario of no project/uuid.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")

    with pytest.raises(AssertionError, match="Project cannot be None"):
        async for _ in distiller_client.request("query", {"query": "test_query"}):
            pass


@pytest.mark.asyncio
async def test_request_with_custom_agent(request, mock_requests_post):
    """
    Test 'request' handles custom agents when status is 'wait'.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    async def mock_custom_agent(query):
        return "processed_query"

    mock_websocket.return_value.recv = AsyncMock(
        side_effect=['{"type": "PING"}', '{"type": "data", "value": "test_message"}']
    )

    await distiller_client.connect(
        project="test_project",
        uuid="test_uuid",
        custom_agent_gallery={"assistant": mock_custom_agent},
        project_version="0",
    )

    distiller_client.executor_dict = {"assistant": AsyncMock()}

    distiller_client.recv = AsyncMock(
        side_effect=[
            {"status": "wait", "role": "assistant", "query": "test_query"},
            {"status": "complete"},
        ]
    )

    distiller_client.send = AsyncMock()
    responses = []
    async for response in distiller_client.request("query", {"query": "test_query"}):
        responses.append(response)
    assert len(responses) == 0
    distiller_client.send.assert_called()


@pytest.mark.asyncio
async def test_request_tool_use_scenario(request, mock_requests_post):
    """
    Test 'request' method handles tool-use scenarios.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    async def mock_tool_agent(**kwargs):
        return "tool_result"

    mock_websocket.return_value.recv = AsyncMock(
        side_effect=['{"type": "PING"}', '{"type": "data", "value": "test_message"}']
    )

    await distiller_client.connect(
        project="test_project",
        uuid="test_uuid",
        project_version="0",
        custom_agent_gallery={"assistant": mock_tool_agent},
    )
    distiller_client.executor_dict = {"assistant": AsyncMock()}

    distiller_client.recv = AsyncMock(
        side_effect=[
            {
                "status": "wait",
                "role": "assistant",
                "tool_use": True,
                "kwargs": {"param": "value"},
            },
            {"status": "complete"},
        ]
    )

    distiller_client.send = AsyncMock()
    responses = []
    async for response in distiller_client.request("query", {"query": "test_query"}):
        responses.append(response)
    assert len(responses) == 0
    distiller_client.send.assert_called()


@pytest.mark.asyncio
async def test_request_invalid_status(request):
    """
    Test 'request' handles invalid status payload.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.return_value.recv.side_effect = [
        {"status": "unknown", "role": "assistant"},
    ]
    distiller_client.project = "test_project"
    distiller_client.uuid = "test_uuid"

    with pytest.raises(Exception):
        await distiller_client.request("query", {"query": "test_query"})


@pytest.mark.asyncio
async def test_query(request):
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.return_value.recv.side_effect = [
        {"status": "wait", "role": "assistant", "query": "test"},
        {"status": "complete"},
    ]
    distiller_client.project = "test_project"
    distiller_client.uuid = "test_uuid"

    distiller_client.send_queue = Queue()
    await distiller_client.send_queue.put({"mock": "payload"})
    distiller_client.receive_queue = Queue()
    await distiller_client.receive_queue.put({"status": "complete"})

    responses = []
    async for response in await distiller_client.query(
        "query", {"query": "test_query"}
    ):
        responses.append(response)
    assert len(responses) == 0


@pytest.mark.asyncio
async def test_retrieve_memory(request):
    """
    Test 'retrieve_memory' calls 'request' with correct arguments.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.return_value.recv.side_effect = [
        {"status": "wait", "role": "assistant", "query": "test_memory_query"},
        {"status": "complete", "role": "memory", "content": "retrieved_memory"},
    ]
    distiller_client.project = "test_project"
    distiller_client.uuid = "test_uuid"

    async def mock_request(*_, **__):
        yield {"status": "wait", "role": "assistant", "query": "test_memory_query"}
        yield {"status": "complete", "role": "memory", "content": "retrieved_memory"}

    distiller_client.request = mock_request
    memory = await distiller_client.retrieve_memory()
    assert memory == "retrieved_memory"


@pytest.mark.asyncio
async def test_add_memory(request):
    """
    Test 'add_memory' calls 'request' with correct arguments.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.return_value.recv.side_effect = [
        {"status": "wait", "role": "assistant", "query": "add_memory"},
        {"status": "complete"},
    ]
    distiller_client.project = "test_project"
    distiller_client.uuid = "test_uuid"

    async def mock_request(*_, **__):
        yield {"status": "wait", "role": "assistant", "query": "add_memory"}
        yield {"status": "complete"}

    distiller_client.request = mock_request
    await distiller_client.add_memory(test_key="test_value")


@pytest.mark.asyncio
async def test_reset_memory(request):
    """
    Test 'reset_memory' calls 'request' with correct arguments.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.return_value.recv.side_effect = [
        {"status": "complete"},
    ]
    distiller_client.project = "test_project"
    distiller_client.uuid = "test_uuid"

    async def mock_request(*_, **__):
        yield {"status": "complete"}

    distiller_client.request = mock_request
    await distiller_client.reset_memory()


@pytest.mark.asyncio
async def test_retrieve_history_as_list(request):
    """
    Test 'retrieve_history' returns a list of messages when 'as_string' is False.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_db_client = request.getfixturevalue("fx_mock_db_client")

    mock_db_client.execute_query.return_value = (
        [
            ('{"uuid_timestamp": "2", "content": "message2"}',),
            ('{"uuid_timestamp": "1", "content": "message1"}',),
        ],
        True,
    )

    result = await distiller_client.retrieve_history(
        db_client=mock_db_client,
        project="test_project",
        uuid="test_uuid",
        n_messages=2,
        as_string=False,
    )
    expected_output = [
        {"uuid_timestamp": "2", "content": "message2"},
        {"uuid_timestamp": "1", "content": "message1"},
    ]
    assert result == expected_output


@pytest.mark.asyncio
async def test_retrieve_history_as_string(request):
    """
    Test 'retrieve_history' returns a formatted string when 'as_string' is True.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_db_client = request.getfixturevalue("fx_mock_db_client")

    mock_db_client.execute_query.return_value = (
        [
            ('{"uuid_timestamp": "1", "content": "message1"}',),
            ('{"uuid_timestamp": "2", "content": "message2"}',),
        ],
        True,
    )

    result = await distiller_client.retrieve_history(
        db_client=mock_db_client,
        project="test_project",
        uuid="test_uuid",
        n_messages=2,
        as_string=True,
    )

    expected_output = (
        'JSONSTART{"uuid_timestamp": "2", "content": "message2"}JSONEND'
        'JSONSTART{"uuid_timestamp": "1", "content": "message1"}JSONEND'
    )
    assert result == expected_output
    mock_db_client.execute_query.assert_called_once()


@pytest.mark.asyncio
async def test_retrieve_history_failure(request):
    """
    Test a database query failure with empty return value.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_db_client = request.getfixturevalue("fx_mock_db_client")

    mock_db_client.execute_query.return_value = (None, False)
    history = await distiller_client.retrieve_history(
        db_client=mock_db_client,
        project="test_project",
        uuid="test_uuid",
        n_messages=1,
        as_string=False,
    )
    assert history == []


@pytest.mark.asyncio
async def test_retrieve_history_json_decode_error(request):
    """
    Test that 'retrieve_history' raises JSONDecodeError when data is invalid JSON.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_db_client = request.getfixturevalue("fx_mock_db_client")

    mock_db_client.execute_query.return_value = ([("invalid json",)], True)

    with pytest.raises(json.JSONDecodeError):
        await distiller_client.retrieve_history(
            db_client=mock_db_client,
            project="test_project",
            uuid="test_uuid",
            n_messages=5,
        )


@pytest.mark.asyncio
async def test_log_chat_success(request):
    """
    Test that '_log_chat' successfully logs a message to the database.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_db_client = request.getfixturevalue("fx_mock_db_client")

    # pylint: disable=protected-access
    mock_db_client.execute_query.side_effect = [("mock", True), ("mock", True)]
    message = {
        "uuid_timestamp": "test-123",
        "timestamp": 123456789.0,
        "role": "assistant",
        "content": "test content",
    }
    result = await distiller_client._log_chat(
        db_client=mock_db_client,
        project="test_project",
        uuid="test_uuid",
        message=message,
    )
    assert result is True
    assert mock_db_client.execute_query.call_count == 2


@pytest.mark.asyncio
async def test_log_chat_failure(request):
    """
    Test that '_log_chat' returns False when logging fails.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_db_client = request.getfixturevalue("fx_mock_db_client")

    # pylint: disable=protected-access
    mock_db_client.execute_query.side_effect = [("mock", True), ("mock", False)]
    message = {
        "uuid_timestamp": "test-123",
        "timestamp": 123456789.0,
        "role": "assistant",
        "content": "test content",
    }
    result = await distiller_client._log_chat(
        db_client=mock_db_client,
        project="test_project",
        uuid="test_uuid",
        message=message,
    )
    assert result is False
    assert mock_db_client.execute_query.call_count == 2


@pytest.mark.asyncio
async def test_context_manager(request, mock_requests_post):
    """
    Test that AsyncDistillerClient can be used as an async context manager.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    # pylint: disable=protected-access
    mock_websocket.return_value.recv = AsyncMock(
        side_effect=['{"type": "PING"}', '{"type": "data", "value": "test_message"}']
    )
    async with distiller_client(project="test_project", uuid="test_uuid") as ctx_client:
        assert ctx_client.connection is not None
        assert ctx_client._send_task is not None
        assert ctx_client._receive_task is not None

    assert distiller_client.connection is None


@pytest.mark.asyncio
async def test_interactive_helper(fx_distiller_client, fx_mock_websocket):
    """
    Test the _interactive_helper method for handling interactive mode.
    """
    with patch("air.distiller.client.async_input") as mock_input, patch(
        "air.distiller.client.async_print"
    ) as mock_print:

        mock_input.side_effect = [
            "Hello!",  # First user input
            KeyboardInterrupt,  # To break out of the infinite loop
        ]

        mock_response = {"role": "assistant", "content": "Hello! How can I help you?"}
        fx_distiller_client.query = AsyncMock()

        async def mock_query_responses():
            yield mock_response

        fx_distiller_client.query.return_value = mock_query_responses()
        fx_distiller_client.connect = AsyncMock()
        fx_distiller_client.close = AsyncMock()

        try:
            await fx_distiller_client._interactive_helper(
                project="test-project", uuid="test-uuid", executor_dict={}
            )
        except KeyboardInterrupt:
            pass

        # Verify connect was called with the correct arguments
        fx_distiller_client.connect.assert_called_once_with(
            project="test-project", uuid="test-uuid", executor_dict={}
        )
        fx_distiller_client.query.assert_called_once_with("Hello!")
        mock_input.assert_called_with("%%% USER %%%\n")

        # Verify async_print was called to output the agent's response
        # Should be called at least 3 times: for the newline, role, and content
        assert mock_print.call_count >= 3

        # Check specific print calls
        mock_print_calls = mock_print.call_args_list
        assert any(call[0] == ("%%% AGENT assistant %%%",) for call in mock_print_calls)
        assert any(
            call[0] == ("Hello! How can I help you?",) for call in mock_print_calls
        )


def test_interactive_method(request):
    """
    Test that 'interactive' calls '_interactive_helper'.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    loop = asyncio.new_event_loop()

    with patch("asyncio.get_event_loop", return_value=loop), patch(
        "asyncio.new_event_loop", return_value=loop
    ), patch.object(loop, "run_until_complete") as mock_run, patch.object(
        distiller_client, "_interactive_helper"
    ) as mock_helper:
        distiller_client.interactive(project="test_project", uuid="test_uuid")
        mock_run.assert_called_once()
        mock_helper.assert_called_once_with("test_project", "test_uuid", None)


def test_create_project_with_json_config(request):
    """
    Test that create_project works with json_config instead of config_path.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_response = MagicMock()
    mock_response.status_code = 201
    mock_response.content = json.dumps({"project_version": -1})
    json_config = {"model": "gpt-4", "temperature": 0.7}

    with patch("requests.post", return_value=mock_response) as mock_post:
        result = distiller_client.create_project(
            project="test_project", json_config=json_config
        )
        assert result is True
        mock_post.assert_called_once()


def test_create_project_no_config(request):
    """
    Test that create_project raises an exception if neither config_path nor json_config is provided.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    with pytest.raises(
        Exception, match="Either json_config or config_path must be provided."
    ):
        distiller_client.create_project(project="test_project")


@pytest.mark.asyncio
async def test_connect_websocket_version_handling(request, mock_requests_post):
    """
    Test that connect handles different websocket versions correctly.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")
    mock_connection = mock_websocket.return_value  # Access the AsyncMock connection

    # Use MagicMock instead of AsyncMock for version
    with patch("air.distiller.client.version", MagicMock(return_value="13.0")):
        # Set side effects on the 'recv' method
        mock_connection.recv.side_effect = [
            '{"type": "PING"}',
            '{"type": "data", "value": "test"}',
        ]
        await distiller_client.connect(
            project="test_project",
            uuid="test_uuid",
            project_version="0",
        )
        assert distiller_client.connection is not None
        await distiller_client.close()  # Make sure to properly close


@pytest.mark.asyncio
async def test_receive_non_json_message(request, mock_requests_post):
    """
    Test handling of non-JSON messages in receive loop.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.return_value.recv = AsyncMock(
        side_effect=["invalid json", '{"type": "PING"}']
    )

    await distiller_client.connect(
        project="test_project",
        uuid="test_uuid",
        project_version="0",
    )
    await asyncio.sleep(0.1)  # Let the receive loop process
    assert distiller_client.connection is not None


@pytest.mark.asyncio
async def test_close_with_timeout(request, mock_requests_post):
    """
    Test close method when tasks timeout.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.return_value.recv = AsyncMock(
        side_effect=['{"type": "PING"}', '{"type": "data", "value": "test"}']
    )

    await distiller_client.connect(
        project="test_project",
        uuid="test_uuid",
        project_version="0",
    )

    # pylint: disable=protected-access
    async def never_ending_task():
        while True:
            await asyncio.sleep(1)

    distiller_client._wait_task_list = [asyncio.create_task(never_ending_task())]

    await distiller_client.close()
    assert distiller_client.connection is None


@pytest.mark.asyncio
async def test_custom_agent_multiple_params(request, mock_requests_post):
    """
    Test custom agent handling with multiple parameters.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    async def mock_custom_agent(query, param1, param2):
        return f"Processed: {query}, {param1}, {param2}"

    mock_websocket.return_value.recv = AsyncMock(
        side_effect=['{"type": "PING"}', '{"type": "data", "value": "test_message"}']
    )

    await distiller_client.connect(
        project="test_project",
        project_version="0",
        uuid="test_uuid",
        custom_agent_gallery={"assistant": mock_custom_agent},
    )

    distiller_client.executor_dict = {"assistant": AsyncMock()}

    distiller_client.recv = AsyncMock(
        side_effect=[
            {
                "status": "wait",
                "role": "assistant",
                "query": "test_query",
                "kwargs": {"param1": "value1", "param2": "value2"},
            },
            {"status": "complete"},
        ]
    )

    async for _ in distiller_client.request("query", {"query": "test"}):
        pass


@pytest.mark.asyncio
async def test_send_queue_handling(request, mock_requests_post):
    """
    Test send queue handling, including sending None to trigger shutdown.
    """
    distiller_client = request.getfixturevalue("fx_distiller_client")
    mock_websocket = request.getfixturevalue("fx_mock_websocket")

    mock_websocket.return_value.recv = AsyncMock(
        side_effect=['{"type": "PING"}', '{"type": "data", "value": "test"}']
    )

    await distiller_client.connect(
        project="test_project",
        uuid="test_uuid",
        project_version="0",
    )
    await distiller_client.send({"type": "test"})

    # Put None in the send queue to signal the send loop to terminate
    await distiller_client.send_queue.put(None)
    await asyncio.sleep(0.1)  # Give time for the send loop to process

    assert distiller_client.connection is not None
