# pylint: disable=missing-function-docstring,import-outside-toplevel

import pytest
import requests
from unittest.mock import patch, MagicMock
from typing import cast
from aiohttp.typedefs import LooseHeaders

from air.embeddings.client import EmbeddingsClient, AsyncEmbeddingsClient
from air.types import CreateEmbeddingResponse


class MockAiohttpResponse:
    """
    A minimal mock for aiohttp's response, usable as:
       async with session.post(...) as resp:
           ...
    """

    def __init__(self, status=200, json_data=None, raise_for_status=False):
        self.status = status
        self._json_data = json_data or {}
        self._raise_for_status = raise_for_status

    async def __aenter__(self):
        if self._raise_for_status and self.status >= 400:
            from aiohttp import ClientResponseError

            # Provide a minimal but valid RequestInfo
            from aiohttp.client_reqrep import RequestInfo
            from yarl import URL

            request_info = RequestInfo(
                url=URL("http://dummy-url"),
                method="GET",
                headers={},  # type: ignore[arg-type]
                real_url=URL("http://dummy-url"),
            )
            raise ClientResponseError(
                request_info=request_info,
                history=(),
                status=self.status,
                message="Mocked Error",
            )
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def json(self):
        return self._json_data

    def raise_for_status(self):
        if self._raise_for_status and self.status >= 400:
            from aiohttp import ClientResponseError
            from aiohttp.client_reqrep import RequestInfo
            from yarl import URL

            request_info = RequestInfo(
                url=URL("http://dummy-url"),
                method="GET",
                headers={},  # type: ignore[arg-type]
                real_url=URL("http://dummy-url"),
            )
            raise ClientResponseError(
                request_info=request_info,
                history=(),
                status=self.status,
                message="Mocked Error",
            )


@pytest.fixture
def sample_create_embedding_response():
    return {
        "object": "list",
        "model": "intfloat/e5-mistral-7b-instruct",
        "data": [
            {"object": "embedding", "embedding": [0.123, -0.456, 0.789], "index": 0}
        ],
        "usage": {"prompt_tokens": 5, "total_tokens": 5},
    }


@pytest.mark.asyncio
class TestAsyncEmbeddingsClient:
    async def test_create_sends_correct_request(self, sample_create_embedding_response):
        with patch("aiohttp.ClientSession", autospec=True) as mock_session_cls:
            mock_session = MagicMock()
            mock_session_cls.return_value.__aenter__.return_value = mock_session

            # .post(...) should return a context manager supporting async __aenter__/__aexit__
            mock_session.post.return_value = MockAiohttpResponse(
                status=200, json_data=sample_create_embedding_response
            )

            client = AsyncEmbeddingsClient(
                base_url="https://example.com", api_key="fake_key"
            )
            resp = await client.create(
                model="intfloat/e5-mistral-7b-instruct",
                input=["async text"],
                user="testing-async",
            )

            # Cast headers dict to LooseHeaders to make Pyright happy
            expected_headers = cast(
                LooseHeaders,
                {
                    "Authorization": "Bearer fake_key",
                    "Content-Type": "application/json",
                },
            )

            mock_session.post.assert_called_once_with(
                "https://example.com/embeddings",
                json={
                    "model": "intfloat/e5-mistral-7b-instruct",
                    "input": ["async text"],
                    "user": "testing-async",
                    "extra_body": None,
                },
                headers=expected_headers,
            )

            assert isinstance(resp, CreateEmbeddingResponse)
            assert len(resp.data) == 1
            assert resp.data[0].embedding == [0.123, -0.456, 0.789]

    async def test_create_raises_for_status(self):
        with patch("aiohttp.ClientSession", autospec=True) as mock_session_cls:
            mock_session = MagicMock()
            mock_session_cls.return_value.__aenter__.return_value = mock_session

            # Return a 400 and raise_for_status -> triggers the exception
            mock_session.post.return_value = MockAiohttpResponse(
                status=400,
                json_data={"error": "Something went wrong"},
                raise_for_status=True,
            )

            client = AsyncEmbeddingsClient(
                base_url="https://example.com", api_key="fake_key"
            )

            from aiohttp import ClientResponseError

            with pytest.raises(ClientResponseError):
                await client.create(model="some-model", input=["fail me"])


class TestEmbeddingsClient:
    def test_create_sends_correct_request(
        self, mocker, sample_create_embedding_response
    ):
        mock_post = mocker.patch("requests.post")
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = sample_create_embedding_response

        client = EmbeddingsClient(base_url="https://example.com", api_key="fake_key")
        resp = client.create(model="test-model", input=["a", "b"], user="xyz")

        mock_post.assert_called_once_with(
            "https://example.com/embeddings",
            json={
                "model": "test-model",
                "input": ["a", "b"],
                "user": "xyz",
                "extra_body": None,
            },
            headers={
                "Authorization": "Bearer fake_key",
                "Content-Type": "application/json",
            },
            timeout=60,  # default
        )
        assert isinstance(resp, CreateEmbeddingResponse)
        assert resp.data[0].embedding == [0.123, -0.456, 0.789]

    def test_create_raises_for_status(self, mocker):
        mock_post = mocker.patch("requests.post")
        mock_resp = MagicMock()
        mock_resp.raise_for_status.side_effect = requests.HTTPError("Bad Request")
        mock_post.return_value = mock_resp

        client = EmbeddingsClient(base_url="https://example.com", api_key="fake_key")
        with pytest.raises(requests.HTTPError):
            client.create(model="test-model", input=["fail"])
