import unittest
from unittest.mock import patch, MagicMock, AsyncMock

from air.images.client import ImagesClient, AsyncImagesClient
from air.types.images import ImagesResponse


class TestImagesClient(unittest.TestCase):
    """Tests for the synchronous ImagesClient."""

    def setUp(self):
        """Sets up common attributes for each test."""
        self.base_url = "https://api.airefinery.accenture.com"
        self.api_key = "test_api_key"
        self.client = ImagesClient(base_url=self.base_url, api_key=self.api_key)

    @patch("requests.post")
    def test_generate_successful(self, mock_post):
        """Tests that a successful call returns a valid ImagesResponse.

        This test verifies:
          * A correct POST request is sent to the /images/generations endpoint.
          * The return value is a properly constructed ImagesResponse containing
            expected fields and data.
        """
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "created": 1697047400,
            "data": [
                {
                    "b64_json": "some_base64_string",
                    "revised_prompt": "revised prompt text",
                    "url": "https://example.com/image.png",
                }
            ],
            "usage": {
                "input_tokens": 10,
                "input_tokens_details": {"type_a": 5, "type_b": 5},
                "output_tokens": 1,
                "total_tokens": 11,
            },
        }
        mock_post.return_value = mock_response

        response = self.client.generate(prompt="A prompt", model="test-model")

        # Check the endpoint was called with the correct URL
        expected_url = f"{self.base_url}/images/generations"
        mock_post.assert_called_once()
        call_args, call_kwargs = mock_post.call_args
        self.assertEqual(call_args[0], expected_url)

        # Check headers
        self.assertIn("Authorization", call_kwargs["headers"])
        self.assertIn("Content-Type", call_kwargs["headers"])
        self.assertEqual(
            call_kwargs["headers"]["Authorization"], f"Bearer {self.api_key}"
        )

        # Check JSON payload
        self.assertEqual(call_kwargs["json"]["prompt"], "A prompt")
        self.assertEqual(call_kwargs["json"]["model"], "test-model")

        # Validate the response
        self.assertIsInstance(response, ImagesResponse)
        self.assertEqual(response.created, 1697047400)
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0].url, "https://example.com/image.png")

    @patch("requests.post")
    def test_generate_raises_for_http_error(self, mock_post):
        """Tests that an HTTP error is properly raised as an exception."""
        mock_response = MagicMock()
        mock_response.raise_for_status.side_effect = Exception("HTTP Error")
        mock_post.return_value = mock_response

        with self.assertRaises(Exception):
            self.client.generate(prompt="Bad prompt", model="test-model")


class TestAsyncImagesClient(unittest.IsolatedAsyncioTestCase):
    """Tests for the asynchronous AsyncImagesClient."""

    def setUp(self):
        """Sets up common attributes for each async test."""
        self.base_url = "https://api.airefinery.accenture.com"
        self.api_key = "test_api_key"
        self.async_client = AsyncImagesClient(
            base_url=self.base_url, api_key=self.api_key
        )

    @patch("aiohttp.ClientSession.post")
    async def test_generate_successful_async(self, mock_post):
        """Tests that a successful async call returns a valid ImagesResponse.

        This test verifies:
          * A correct async POST request is sent to the /images/generations endpoint.
          * The return value is a properly constructed ImagesResponse containing
            expected fields and data.
        """
        mock_resp = MagicMock()
        mock_resp.status = 200

        mock_resp.json = AsyncMock(
            return_value={
                "created": 1697047400,
                "data": [
                    {
                        "b64_json": "some_base64_string",
                        "revised_prompt": "revised prompt text",
                        "url": "https://example.com/image.png",
                    }
                ],
                "usage": {
                    "input_tokens": 10,
                    "input_tokens_details": {"type_a": 5, "type_b": 5},
                    "output_tokens": 1,
                    "total_tokens": 11,
                },
            }
        )

        mock_cm = MagicMock()
        mock_cm.__aenter__.return_value = mock_resp
        mock_post.return_value = mock_cm

        response = await self.async_client.generate(
            prompt="A prompt", model="test-model"
        )

        # Verify the call was made to the correct endpoint
        expected_url = f"{self.base_url}/images/generations"
        mock_post.assert_called_once()
        call_args, call_kwargs = mock_post.call_args
        self.assertEqual(call_args[0], expected_url)

        # Check headers
        self.assertIn("Authorization", call_kwargs["headers"])
        self.assertIn("Content-Type", call_kwargs["headers"])
        self.assertEqual(
            call_kwargs["headers"]["Authorization"], f"Bearer {self.api_key}"
        )

        # Check JSON payload
        self.assertEqual(call_kwargs["json"]["prompt"], "A prompt")
        self.assertEqual(call_kwargs["json"]["model"], "test-model")

        # Validate the response
        self.assertIsInstance(response, ImagesResponse)
        self.assertEqual(response.created, 1697047400)
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0].url, "https://example.com/image.png")

    @patch("aiohttp.ClientSession.post")
    async def test_generate_raises_for_http_error_async(self, mock_post):
        """Tests that an HTTP error in the async client raises an exception."""
        mock_resp = MagicMock()
        mock_resp.raise_for_status.side_effect = Exception("Async HTTP Error")

        mock_cm = MagicMock()
        mock_cm.__aenter__.return_value = mock_resp
        mock_post.return_value = mock_cm

        with self.assertRaises(Exception):
            await self.async_client.generate(prompt="Bad prompt", model="test-model")
