import pytest
import requests
import aiohttp

from unittest.mock import patch, MagicMock

# Adjust if your imports differ
from air.models.client import ModelsClient, AsyncModelsClient
from air.types import SyncPage, AsyncPage
from air.types.models import Model


@pytest.fixture
def sample_models_page_response():
    """
    Example JSON matching your SyncPage[Model] or AsyncPage[Model] structure.
    Adjust as needed for your real API responses.
    """
    return {
        "object": "list",
        "data": [
            {
                "id": "gpt-4o-mini",
                "created": 1234567890,
                "object": "model",
                "owned_by": "openai",
                "root": "gpt-4o-mini",
                "parent": None,
                "permission": [
                    {
                        "id": "modelperm-xyz",
                        "object": "model_permission",
                        "created": 1234567890,
                        "allow_create_engine": True,
                        "allow_sampling": True,
                        "allow_logprobs": True,
                        "allow_search_indices": True,
                        "allow_view": True,
                        "allow_fine_tuning": False,
                        "organization": "openai",
                        "group": None,
                        "is_blocking": False,
                    }
                ],
            }
        ],
    }


class TestModelsClient:
    """
    Tests for the synchronous ModelsClient.list(...) method.
    """

    def test_list_sends_correct_request(self, mocker, sample_models_page_response):
        """
        Patches requests.get() so we can verify the correct endpoint, headers,
        JSON payload, etc.
        """
        mock_get = mocker.patch("requests.get")
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = sample_models_page_response

        client = ModelsClient(base_url="https://test.com", api_key="fake_key")

        # Call the method under test
        resp = client.list(timeout=10.0, some_param="value")

        # Check requests.get() was called as expected
        mock_get.assert_called_once_with(
            "https://test.com/models",
            json={"some_param": "value"},  # from **kwargs
            headers={
                "Authorization": "Bearer fake_key",
                "Content-Type": "application/json",
            },
            timeout=10.0,
        )

        # The response should be parsed into a SyncPage[Model]
        assert isinstance(resp, SyncPage)
        assert len(resp.data) == 1
        assert isinstance(resp.data[0], Model)
        assert resp.data[0].id == "gpt-4o-mini"

    def test_list_raises_for_status(self, mocker):
        """Ensures an HTTP error triggers requests.HTTPError."""
        mock_get = mocker.patch("requests.get")
        mock_resp = MagicMock()
        mock_resp.raise_for_status.side_effect = requests.HTTPError("Bad Request")
        mock_get.return_value = mock_resp

        client = ModelsClient(base_url="https://test.com", api_key="fake_key")

        with pytest.raises(requests.HTTPError):
            client.list()


@pytest.mark.asyncio
class TestAsyncModelsClient:
    """
    Tests for the asynchronous AsyncModelsClient.list(...) method.
    """

    async def test_list_sends_correct_request(self, sample_models_page_response):
        with patch("aiohttp.ClientSession", autospec=True) as mock_session_cls:
            # Mock the session returned by "with ClientSession():"
            mock_session = MagicMock()
            mock_session_cls.return_value.__aenter__.return_value = mock_session

            # We'll define a small mock for "async with session.get(...)".
            class MockAiohttpResponse:
                def __init__(self, status=200, json_data=None):
                    self.status = status
                    self._json_data = json_data or {}

                async def __aenter__(self):
                    return self

                async def __aexit__(self, exc_type, exc_val, exc_tb):
                    pass

                async def json(self):
                    return self._json_data

                def raise_for_status(self):
                    if self.status >= 400:
                        from aiohttp.client_reqrep import RequestInfo
                        from yarl import URL

                        request_info = RequestInfo(
                            url=URL("http://dummy-url"),
                            method="GET",
                            headers={},  # type: ignore[arg-type]
                            real_url=URL("http://dummy-url"),
                        )
                        raise aiohttp.ClientResponseError(
                            request_info=request_info,
                            history=(),
                            status=self.status,
                            message="Mocked Error",
                        )

            # Return a successful 200 JSON response
            mock_session.get.return_value = MockAiohttpResponse(
                status=200, json_data=sample_models_page_response
            )

            client = AsyncModelsClient(base_url="https://test.com", api_key="fake_key")

            # Call the method under test
            resp = await client.list(timeout=15.0, alt="stuff")

            # Verify the request was made properly
            mock_session.get.assert_called_once_with(
                "https://test.com/models",
                json={"alt": "stuff"},
                headers={
                    "Authorization": "Bearer fake_key",
                    "Content-Type": "application/json",
                },
            )

            # The response should be parsed into an AsyncPage[Model]
            assert isinstance(resp, AsyncPage)
            assert len(resp.data) == 1
            assert isinstance(resp.data[0], Model)
            assert resp.data[0].id == "gpt-4o-mini"

    async def test_list_raises_for_status(self):
        """
        Mocks a 400 response to confirm raise_for_status triggers aiohttp.ClientResponseError.
        """
        with patch("aiohttp.ClientSession", autospec=True) as mock_session_cls:
            mock_session = MagicMock()
            mock_session_cls.return_value.__aenter__.return_value = mock_session

            class MockAiohttpResponse:
                def __init__(self, status=400):
                    self.status = status

                async def __aenter__(self):
                    return self

                async def __aexit__(self, exc_type, exc_val, exc_tb):
                    pass

                async def json(self):
                    return {"error": "Something went wrong"}

                def raise_for_status(self):
                    if self.status >= 400:
                        from aiohttp.client_reqrep import RequestInfo
                        from yarl import URL

                        request_info = RequestInfo(
                            url=URL("http://dummy-url"),
                            method="GET",
                            headers={},  # type: ignore[arg-type]
                            real_url=URL("http://dummy-url"),
                        )
                        raise aiohttp.ClientResponseError(
                            request_info=request_info,
                            history=(),
                            status=self.status,
                            message="Mocked Error",
                        )

            mock_session.get.return_value = MockAiohttpResponse(status=400)

            client = AsyncModelsClient(base_url="https://test.com", api_key="fake_key")

            with pytest.raises(aiohttp.ClientResponseError):
                await client.list()
