# pylint: disable=unused-import, too-many-positional-arguments, unused-argument

import json
import time
import pytest
from unittest.mock import patch, MagicMock
import requests

from air import __token_url__, __base_url__
from air.authenticator import Authenticator  # Adjusted import based on provided module


@pytest.fixture
def valid_credentials():
    """
    Fixture for valid credentials.
    """
    return {
        "account": "test_account",
        "api_key": "test_api_key",
    }


@pytest.fixture
def invalid_credentials():
    """
    Fixture for invalid credentials.
    """
    return {
        "account": "invalid_account",
        "api_key": "invalid_api_key",
    }


@patch("requests.post")
def test_login_success(mock_post, valid_credentials):
    """
    Test the login method for successful authentication.
    """
    # Arrange
    mock_response = MagicMock()
    mock_response.json.return_value = {"access_token": "mock_access_token"}
    mock_response.raise_for_status.return_value = None
    mock_post.return_value = mock_response

    authenticator = Authenticator(
        account=valid_credentials["account"],
        api_key=valid_credentials["api_key"],
    )

    # Act
    access_token = authenticator.get_access_token()

    # Assert
    assert access_token == "mock_access_token"
    mock_post.assert_called_once_with(
        __token_url__,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={
            "client_id": valid_credentials["account"],
            "scope": "https://graph.microsoft.com/.default",
            "client_secret": valid_credentials["api_key"],
            "grant_type": "client_credentials",
        },
        timeout=10,  # Include the timeout as specified in your code
    )


@patch("requests.post")
def test_login_failure(mock_post, invalid_credentials):
    """
    Test the login method for failed authentication.
    """
    # Arrange
    mock_response = MagicMock()
    mock_response.raise_for_status.side_effect = requests.HTTPError("401 Client Error")
    mock_post.return_value = mock_response

    authenticator = Authenticator(
        account=invalid_credentials["account"],
        api_key=invalid_credentials["api_key"],
    )

    # Act
    access_token = authenticator.get_access_token()

    # Assert
    assert access_token == ""  # Expecting an empty string for failed login
    mock_post.assert_called_once_with(
        __token_url__,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={
            "client_id": invalid_credentials["account"],
            "scope": "https://graph.microsoft.com/.default",
            "client_secret": invalid_credentials["api_key"],
            "grant_type": "client_credentials",
        },
        timeout=10,
    )


@patch("requests.post")
def test_get_access_token_valid(mock_post, valid_credentials):
    """
    Test getting access token when it is still valid.
    """
    # Arrange
    mock_response = MagicMock()
    mock_response.json.return_value = {"access_token": "mock_access_token"}
    mock_response.raise_for_status.return_value = None
    mock_post.return_value = mock_response

    authenticator = Authenticator(
        account=valid_credentials["account"],
        api_key=valid_credentials["api_key"],
    )

    # Simulate time passage less than the timeout
    time.sleep(1)  # Sleep for a second to ensure we are still within timeout
    access_token = authenticator.get_access_token()

    # Assert
    assert access_token == "mock_access_token"  # Token should still be valid
    mock_post.assert_called_once()  # Should only be called once during initialization


@patch("requests.post")
def test_get_access_token_expired(mock_post, valid_credentials):
    """
    Test getting access token when it has expired.
    """
    # Arrange
    mock_response = MagicMock()
    mock_response.json.return_value = {"access_token": "new_mock_access_token"}
    mock_response.raise_for_status.return_value = None
    mock_post.return_value = mock_response

    authenticator = Authenticator(
        account=valid_credentials["account"],
        api_key=valid_credentials["api_key"],
    )

    # Simulate time passage greater than the timeout
    authenticator.time = time.time() - (authenticator.timeout + 1)  # Force expiration

    # Act
    access_token = authenticator.get_access_token()

    # Assert
    assert (
        access_token == "new_mock_access_token"
    )  # Should call login again and get a new token
    assert mock_post.call_count == 2  # It should call login again to get a new token


@patch("requests.post")
def test_openai_with_default_base_url(mock_post, valid_credentials):
    """
    Test the openai method with default base_url (None).
    """
    # Arrange
    mock_response = MagicMock()
    mock_response.json.return_value = {"access_token": "mock_access_token"}
    mock_response.raise_for_status.return_value = None
    mock_post.return_value = mock_response

    authenticator = Authenticator(
        account=valid_credentials["account"],
        api_key=valid_credentials["api_key"],
    )

    # Act
    openai_config = authenticator.openai()

    # Assert
    assert openai_config["base_url"] == f"{__base_url__}/inference"
    assert openai_config["api_key"] == "mock_access_token"
    assert openai_config["default_headers"] == {"airefinery_account": "test_account"}


@patch("requests.post")
def test_openai_with_custom_base_url(mock_post, valid_credentials):
    """
    Test the openai method with a custom base_url.
    """
    # Arrange
    mock_response = MagicMock()
    mock_response.json.return_value = {"access_token": "mock_access_token"}
    mock_response.raise_for_status.return_value = None
    mock_post.return_value = mock_response

    authenticator = Authenticator(
        account=valid_credentials["account"],
        api_key=valid_credentials["api_key"],
    )

    # Act
    openai_config = authenticator.openai(base_url="https://custom_base_url.com")

    # Assert
    assert openai_config["base_url"] == "https://custom_base_url.com/inference"
    assert openai_config["api_key"] == "mock_access_token"
    assert openai_config["default_headers"] == {"airefinery_account": "test_account"}


@patch("requests.post")
def test_openai_invalid_login(mock_post, invalid_credentials):
    """
    Test the openai method behavior with invalid credentials (failed login).
    """
    # Arrange
    mock_response = MagicMock()
    mock_response.raise_for_status.side_effect = requests.HTTPError("401 Client Error")
    mock_post.return_value = mock_response

    authenticator = Authenticator(
        account=invalid_credentials["account"],
        api_key=invalid_credentials["api_key"],
    )

    # Act
    openai_config = authenticator.openai()

    # Assert
    assert openai_config["api_key"] == ""  # Expecting an empty api_key for failed login
    mock_post.assert_called()


@patch("requests.post")
def test_default_oauth_server(mock_post, valid_credentials):
    """
    Test the login method with default OAUTH method.
    """
    # Arrange
    mock_response = MagicMock()
    mock_response.json.return_value = {"access_token": "mock_access_token"}
    mock_response.raise_for_status.return_value = None
    mock_post.return_value = mock_response

    authenticator = Authenticator(
        account=valid_credentials["account"],
        api_key=valid_credentials["api_key"],
    )

    # Act
    access_token = authenticator.get_access_token()
    oauth_server = authenticator.oauth_server

    # Assert
    assert oauth_server == ""
    mock_post.assert_called_once_with(
        __token_url__,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={
            "client_id": valid_credentials["account"],
            "scope": "https://graph.microsoft.com/.default",
            "client_secret": valid_credentials["api_key"],
            "grant_type": "client_credentials",
        },
        timeout=10,  # Include the timeout as specified in your code
    )


@patch("requests.post")
def test_custom_oauth_server(mock_post, valid_credentials):
    """
    Test the login method with a custom OAUTH server.
    """
    # Arrange
    mock_response = MagicMock()
    mock_response.json.return_value = {"access_token": "mock_access_token"}
    mock_response.raise_for_status.return_value = None
    mock_post.return_value = mock_response

    authenticator = Authenticator(
        account=valid_credentials["account"],
        api_key=valid_credentials["api_key"],
        oauth_server="custom_oauth_url",
    )

    # Act
    access_token = authenticator.get_access_token()
    oauth_server = authenticator.oauth_server

    # Assert
    assert oauth_server == "custom_oauth_url"
    mock_post.assert_called_once_with(
        "custom_oauth_url/realms/airefinery-realm/protocol/openid-connect/token",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={
            "client_id": valid_credentials["account"],
            "client_secret": valid_credentials["api_key"],
            "grant_type": "client_credentials",
        },
        timeout=10,  # Include the timeout as specified in your code
    )
