from unittest.mock import patch

import pytest

from air.chat import AsyncChatClient, ChatClient
from air.client import AIRefinery, AsyncAIRefinery
from air.distiller import AsyncDistillerClient
from air.embeddings import AsyncEmbeddingsClient, EmbeddingsClient
from air.images import AsyncImagesClient, ImagesClient
from air.knowledge import AsyncKnowledgeClient, KnowledgeClient
from air.models import AsyncModelsClient, ModelsClient


@pytest.fixture
def mock_distiller_auth():
    """
    Patch the 'auth' object inside 'air.distiller.client' so it has the
    required attributes for the Distiller client.
    """
    with patch("air.distiller.client.auth", autospec=True) as mock_auth:
        # Provide the attributes the distiller client code expects.
        mock_auth.account = "test_account"
        mock_auth.api_key = "test_api_key"
        yield mock_auth  # yield to the test, revert patch after


def test_async_airefinery_init(mock_distiller_auth):
    """
    Verify AsyncAIRefinery initializes all async sub-clients and
    stores parameters as expected. We rely on mock_distiller_auth
    so we don't get AttributeError when creating AsyncDistillerClient.
    """
    client = AsyncAIRefinery(
        api_key="test-key", base_url="http://example.com", extra_arg="some_value"
    )
    # Confirm stored attributes
    assert client.api_key == "test-key"
    assert client.base_url == "http://example.com"
    assert "extra_arg" in client.kwargs
    assert client.kwargs["extra_arg"] == "some_value"

    # Confirm sub-client types
    assert isinstance(client.chat, AsyncChatClient)
    assert isinstance(client.embeddings, AsyncEmbeddingsClient)
    assert isinstance(client.models, AsyncModelsClient)
    assert isinstance(client.distiller, AsyncDistillerClient)
    assert isinstance(client.images, AsyncImagesClient)
    assert isinstance(client.knowledge, AsyncKnowledgeClient)


def test_airefinery_init():
    """
    Verify AIRefinery initializes all sync sub-clients and
    stores parameters as expected. Distiller is not provided in sync mode.
    """
    client = AIRefinery(
        api_key="sync-key", base_url="http://sync.example.com", custom_param="testing"
    )
    # Confirm stored attributes
    assert client.api_key == "sync-key"
    assert client.base_url == "http://sync.example.com"
    assert "custom_param" in client.kwargs
    assert client.kwargs["custom_param"] == "testing"

    # Confirm sub-client types
    assert isinstance(client.chat, ChatClient)
    assert isinstance(client.embeddings, EmbeddingsClient)
    assert isinstance(client.models, ModelsClient)
    assert isinstance(client.images, ImagesClient)
    assert isinstance(client.knowledge, KnowledgeClient)

    # Distiller is not accessible in the sync client
    with pytest.raises(NotImplementedError) as exc_info:
        _ = client.distiller
    assert "Distiller is only available in async mode" in str(exc_info.value)
