"""
Test suite for image_to_base64 function, covering URL and file path inputs,
different image formats, and error cases.
"""

from io import BytesIO
import pytest
from unittest.mock import mock_open, patch, MagicMock
from PIL import Image
import base64
from requests.models import Response  # Import Response from requests.models
from air.utils.image_helper import image_to_base64, save_base64_image


@pytest.fixture
def mock_image():
    """
    Create a simple test image in memory.
    """
    img = Image.new("RGB", (100, 100), color="red")
    img_buffer = BytesIO()
    img.save(img_buffer, format="JPEG")
    img_buffer.seek(0)
    return img_buffer.getvalue()


@pytest.fixture
def mock_base64_string():
    """
    Create a simple test image in memory and return its base64-encoded string.
    """
    img = Image.new("RGB", (100, 100), color="red")
    img_buffer = BytesIO()
    img.save(img_buffer, format="JPEG")
    img_buffer.seek(0)
    # Encode the image to base64
    return base64.b64encode(img_buffer.getvalue()).decode("utf-8")


def test_empty_input():
    """
    Test with empty input.
    """
    assert image_to_base64("") is None


@pytest.mark.parametrize("url_ext", [".jpg", ".jpeg", ".png"])
def test_valid_url_images(url_ext, mock_image):
    """
    Test processing images from valid URLs with different extensions.
    """
    test_url = f"https://example.com/image{url_ext}"

    # Create a real Response object and set its attributes
    mock_response = Response()
    mock_response.status_code = 200
    mock_response._content = mock_image  # Set the raw content of the response

    with patch("requests.get", return_value=mock_response):
        result = image_to_base64(test_url)

        assert result is not None
        assert isinstance(result, str)
        result = result.split(",", 1)[1]
        # Verify it's a valid base64 string
        assert len(result) % 4 == 0
        assert all(
            c in "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
            for c in result.rstrip("=")
        )


def test_invalid_url():
    """
    Test with URL returning non-200 status code.
    """
    test_url = "https://example.com/image.jpg"

    # Mock failed request
    mock_response = MagicMock()
    mock_response.status_code = 404

    with patch("requests.get", return_value=mock_response):
        with pytest.raises(ValueError) as exc_info:
            image_to_base64(test_url)
        assert "Failed to fetch image from URL" in str(exc_info.value)


def test_invalid_url_extension():
    """
    Test URL with unsupported file extension.
    """
    test_url = "https://example.com/image.gif"

    with pytest.raises(ValueError):
        image_to_base64(test_url)


@pytest.mark.parametrize("file_ext", [".jpg", ".jpeg", ".png"])
def test_valid_file_paths(file_ext, mock_image):
    """
    Test processing images from valid file paths with different extensions.
    """
    test_path = f"/path/to/image{file_ext}"

    # Mock file operations
    mock_file = mock_open(read_data=mock_image)

    with patch("builtins.open", mock_file), patch(
        "PIL.Image.open", return_value=Image.open(BytesIO(mock_image))
    ):
        result = image_to_base64(test_path)

        assert result is not None
        assert isinstance(result, str)
        result = result.split(",", 1)[1]
        # Verify it's a valid base64 string
        assert len(result) % 4 == 0
        assert all(
            c in "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
            for c in result.rstrip("=")
        )


def test_invalid_file_extension():
    """
    Test file path with unsupported extension.
    """
    test_path = "/path/to/image.gif"

    with pytest.raises(FileNotFoundError):
        image_to_base64(test_path)


def test_nonexistent_file():
    """
    Test with a file path that doesn't exist.
    """
    test_path = "/path/to/nonexistent.jpg"

    with patch("builtins.open", mock_open()) as mock_file:
        mock_file.side_effect = FileNotFoundError()
        with pytest.raises(FileNotFoundError):
            image_to_base64(test_path)


@pytest.mark.parametrize(
    "url", ["http://example.com/image.jpg", "https://example.com/image.jpg"]
)
def test_url_protocols(url, mock_image):
    """
    Test both HTTP and HTTPS protocols.
    """
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.content = mock_image

    with patch("requests.get", return_value=mock_response):
        result = image_to_base64(url)
        assert result is not None


@pytest.mark.parametrize("extension", ["jpg", "jpeg", "png"])
def test_save_base64_image_supported_extensions(
    tmp_path, mock_base64_string, extension
):
    """Test saving a valid base64 image with supported extensions (.jpg, .jpeg, .png)."""
    output_path = tmp_path / f"test_image.{extension}"
    result = save_base64_image(mock_base64_string, str(output_path))
    assert result is True
    assert output_path.exists()


def test_save_base64_image_invalid_extension(tmp_path, mock_base64_string):
    """Test saving with an invalid file extension, expecting a return of False."""
    output_path = tmp_path / "test_image.bmp"
    result = save_base64_image(mock_base64_string, str(output_path))
    assert result is False
    assert not output_path.exists()


def test_save_base64_image_invalid_base64(tmp_path):
    """Test with an invalid base64 string, expecting a return of False."""
    invalid_base64 = "invalid_base64_string"
    output_path = tmp_path / "test_image.jpeg"
    result = save_base64_image(invalid_base64, str(output_path))
    assert result is False
    assert not output_path.exists()


def test_save_base64_image_empty_string(tmp_path):
    """Test with an empty base64 string, expecting a return of False."""
    empty_base64 = ""
    output_path = tmp_path / "test_image.jpeg"
    result = save_base64_image(empty_base64, str(output_path))
    assert result is False
    assert not output_path.exists()


def test_save_base64_image_corrupted_data(tmp_path, mock_base64_string):
    """Test with corrupted base64 data (truncated), expecting a return of False."""
    corrupted_base64 = mock_base64_string[
        :-10
    ]  # Truncate the base64 string to corrupt it
    output_path = tmp_path / "test_image.jpeg"
    result = save_base64_image(corrupted_base64, str(output_path))
    assert result is False
    assert not output_path.exists()
